@@ -58,7 +58,7 @@ tree lastiddecl;
 
 int doing_objc_thang;
 
-extern tree lookup_interface ();
+extern tree is_class_name ();
 
 extern int yydebug;
 
@@ -93,15 +93,15 @@ int check_newline ();
 static int ignore_escape_flag = 0;
 
 /* C code produced by gperf version 2.5 (GNU C++ version) */
-/* Command-line: gperf -p -j1 -i 1 -g -o -t -N is_reserved_word -k1,3,$ c-parse.gperf  */ 
+/* Command-line: gperf -p -j1 -i 1 -g -o -t -G -N is_reserved_word -k1,3,$ c-parse.gperf  */
 struct resword { char *name; short token; enum rid rid; };
 
-#define TOTAL_KEYWORDS 53
+#define TOTAL_KEYWORDS 79
 #define MIN_WORD_LENGTH 2
-#define MAX_WORD_LENGTH 13
-#define MIN_HASH_VALUE 7
-#define MAX_HASH_VALUE 102
-/* maximum key range = 96, duplicates = 0 */
+#define MAX_WORD_LENGTH 20
+#define MIN_HASH_VALUE 10
+#define MAX_HASH_VALUE 144
+/* maximum key range = 135, duplicates = 0 */
 
 #ifdef __GNUC__
 __inline
@@ -113,19 +113,19 @@ hash (str, len)
 {
   static unsigned char asso_values[] =
     {
-     103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
-     103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
-     103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
-     103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
-     103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
-     103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
-     103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
-     103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
-     103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
-     103, 103, 103, 103, 103,   1, 103,   2,   1,  24,
-       1,   5,  19,  39,  16,  13, 103,   1,  25,   1,
-      34,  34,  24, 103,  13,  12,   1,  45,  24,   7,
-     103, 103,   2, 103, 103, 103, 103, 103,
+     145, 145, 145, 145, 145, 145, 145, 145, 145, 145,
+     145, 145, 145, 145, 145, 145, 145, 145, 145, 145,
+     145, 145, 145, 145, 145, 145, 145, 145, 145, 145,
+     145, 145, 145, 145, 145, 145, 145, 145, 145, 145,
+     145, 145, 145, 145, 145, 145, 145, 145, 145, 145,
+     145, 145, 145, 145, 145, 145, 145, 145, 145, 145,
+     145, 145, 145, 145,  25, 145, 145, 145, 145, 145,
+     145, 145, 145, 145, 145, 145, 145, 145, 145, 145,
+     145, 145, 145, 145, 145, 145, 145, 145, 145, 145,
+     145, 145, 145, 145, 145,   1, 145,  46,   8,  15,
+      61,   6,  36,  48,   3,   5, 145,  18,  63,  25,
+      29,  76,   1, 145,  13,   2,   1,  51,  37,   9,
+       9,   1,   3, 145, 145, 145, 145, 145,
     };
   register int hval = len;
 
@@ -141,6 +141,119 @@ hash (str, len)
   return hval + asso_values[str[len - 1]];
 }
 
+static struct resword wordlist[] =
+{
+  {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, 
+  {"",}, 
+  {"int",  TYPESPEC, RID_INT},
+  {"",}, {"",}, 
+  {"__typeof__",  TYPEOF, NORID},
+  {"__signed__",  TYPESPEC, RID_SIGNED},
+  {"__imag__",  IMAGPART, NORID},
+  {"switch",  SWITCH, NORID},
+  {"__inline__",  SCSPEC, RID_INLINE},
+  {"else",  ELSE, NORID},
+  {"__iterator__",  SCSPEC, RID_ITERATOR},
+  {"__inline",  SCSPEC, RID_INLINE},
+  {"__extension__",  EXTENSION, NORID},
+  {"struct",  STRUCT, NORID},
+  {"__real__",  REALPART, NORID},
+  {"__const",  TYPE_QUAL, RID_CONST},
+  {"while",  WHILE, NORID},
+  {"__const__",  TYPE_QUAL, RID_CONST},
+  {"case",  CASE, NORID},
+  {"__complex__",  TYPESPEC, RID_COMPLEX},
+  {"__iterator",  SCSPEC, RID_ITERATOR},
+  {"bycopy",  TYPE_QUAL, RID_BYCOPY},
+  {"",}, {"",}, {"",}, 
+  {"__complex",  TYPESPEC, RID_COMPLEX},
+  {"",}, 
+  {"in",  TYPE_QUAL, RID_IN},
+  {"break",  BREAK, NORID},
+  {"@defs",  DEFS, NORID},
+  {"",}, {"",}, {"",}, 
+  {"extern",  SCSPEC, RID_EXTERN},
+  {"if",  IF, NORID},
+  {"typeof",  TYPEOF, NORID},
+  {"typedef",  SCSPEC, RID_TYPEDEF},
+  {"__typeof",  TYPEOF, NORID},
+  {"sizeof",  SIZEOF, NORID},
+  {"",}, 
+  {"return",  RETURN, NORID},
+  {"const",  TYPE_QUAL, RID_CONST},
+  {"__volatile__",  TYPE_QUAL, RID_VOLATILE},
+  {"@private",  PRIVATE, NORID},
+  {"@selector",  SELECTOR, NORID},
+  {"__volatile",  TYPE_QUAL, RID_VOLATILE},
+  {"__asm__",  ASM_KEYWORD, NORID},
+  {"",}, {"",}, 
+  {"continue",  CONTINUE, NORID},
+  {"__alignof__",  ALIGNOF, NORID},
+  {"__imag",  IMAGPART, NORID},
+  {"__attribute__",  ATTRIBUTE, NORID},
+  {"",}, {"",}, 
+  {"__attribute",  ATTRIBUTE, NORID},
+  {"for",  FOR, NORID},
+  {"",}, 
+  {"@encode",  ENCODE, NORID},
+  {"id",  OBJECTNAME, RID_ID},
+  {"static",  SCSPEC, RID_STATIC},
+  {"@interface",  INTERFACE, NORID},
+  {"",}, 
+  {"__signed",  TYPESPEC, RID_SIGNED},
+  {"",}, 
+  {"__label__",  LABEL, NORID},
+  {"",}, {"",}, 
+  {"__asm",  ASM_KEYWORD, NORID},
+  {"char",  TYPESPEC, RID_CHAR},
+  {"",}, 
+  {"inline",  SCSPEC, RID_INLINE},
+  {"out",  TYPE_QUAL, RID_OUT},
+  {"register",  SCSPEC, RID_REGISTER},
+  {"__real",  REALPART, NORID},
+  {"short",  TYPESPEC, RID_SHORT},
+  {"",}, 
+  {"enum",  ENUM, NORID},
+  {"inout",  TYPE_QUAL, RID_INOUT},
+  {"",}, 
+  {"oneway",  TYPE_QUAL, RID_ONEWAY},
+  {"union",  UNION, NORID},
+  {"",}, 
+  {"__alignof",  ALIGNOF, NORID},
+  {"",}, 
+  {"@implementation",  IMPLEMENTATION, NORID},
+  {"",}, 
+  {"@class",  CLASS, NORID},
+  {"",}, 
+  {"@public",  PUBLIC, NORID},
+  {"asm",  ASM_KEYWORD, NORID},
+  {"",}, {"",}, {"",}, {"",}, {"",}, 
+  {"default",  DEFAULT, NORID},
+  {"",}, 
+  {"void",  TYPESPEC, RID_VOID},
+  {"",}, 
+  {"@protected",  PROTECTED, NORID},
+  {"@protocol",  PROTOCOL, NORID},
+  {"",}, {"",}, {"",}, 
+  {"volatile",  TYPE_QUAL, RID_VOLATILE},
+  {"",}, {"",}, 
+  {"signed",  TYPESPEC, RID_SIGNED},
+  {"float",  TYPESPEC, RID_FLOAT},
+  {"@end",  END, NORID},
+  {"",}, {"",}, 
+  {"unsigned",  TYPESPEC, RID_UNSIGNED},
+  {"@compatibility_alias",  ALIAS, NORID},
+  {"double",  TYPESPEC, RID_DOUBLE},
+  {"",}, {"",}, 
+  {"auto",  SCSPEC, RID_AUTO},
+  {"",}, 
+  {"goto",  GOTO, NORID},
+  {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, 
+  {"do",  DO, NORID},
+  {"",}, {"",}, {"",}, {"",}, 
+  {"long",  TYPESPEC, RID_LONG},
+};
+
 #ifdef __GNUC__
 __inline
 #endif
@@ -149,82 +262,6 @@ is_reserved_word (str, len)
      register char *str;
      register unsigned int len;
 {
-  static struct resword wordlist[] =
-    {
-      {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, 
-      {"asm",  ASM_KEYWORD, NORID},
-      {"",}, 
-      {"__asm",  ASM_KEYWORD, NORID},
-      {"",}, 
-      {"__asm__",  ASM_KEYWORD, NORID},
-      {"break",  BREAK, NORID},
-      {"__typeof__",  TYPEOF, NORID},
-      {"",}, 
-      {"__alignof__",  ALIGNOF, NORID},
-      {"",}, 
-      {"__attribute__",  ATTRIBUTE, NORID},
-      {"int",  TYPESPEC, RID_INT},
-      {"__attribute",  ATTRIBUTE, NORID},
-      {"__extension__",  EXTENSION, NORID},
-      {"",}, 
-      {"__signed",  TYPESPEC, RID_SIGNED},
-      {"",}, 
-      {"__signed__",  TYPESPEC, RID_SIGNED},
-      {"__inline__",  SCSPEC, RID_INLINE},
-      {"else",  ELSE, NORID},
-      {"__inline",  SCSPEC, RID_INLINE},
-      {"default",  DEFAULT, NORID},
-      {"__typeof",  TYPEOF, NORID},
-      {"while",  WHILE, NORID},
-      {"__alignof",  ALIGNOF, NORID},
-      {"struct",  STRUCT, NORID},
-      {"__const",  TYPE_QUAL, RID_CONST},
-      {"if",  IF, NORID},
-      {"__const__",  TYPE_QUAL, RID_CONST},
-      {"__label__",  LABEL, NORID},
-      {"do",  DO, NORID},
-      {"__volatile__",  TYPE_QUAL, RID_VOLATILE},
-      {"sizeof",  SIZEOF, NORID},
-      {"__volatile",  TYPE_QUAL, RID_VOLATILE},
-      {"auto",  SCSPEC, RID_AUTO},
-      {"void",  TYPESPEC, RID_VOID},
-      {"char",  TYPESPEC, RID_CHAR},
-      {"static",  SCSPEC, RID_STATIC},
-      {"case",  CASE, NORID},
-      {"extern",  SCSPEC, RID_EXTERN},
-      {"switch",  SWITCH, NORID},
-      {"for",  FOR, NORID},
-      {"inline",  SCSPEC, RID_INLINE},
-      {"typeof",  TYPEOF, NORID},
-      {"typedef",  SCSPEC, RID_TYPEDEF},
-      {"short",  TYPESPEC, RID_SHORT},
-      {"",}, 
-      {"return",  RETURN, NORID},
-      {"enum",  ENUM, NORID},
-      {"",}, 
-      {"double",  TYPESPEC, RID_DOUBLE},
-      {"signed",  TYPESPEC, RID_SIGNED},
-      {"float",  TYPESPEC, RID_FLOAT},
-      {"",}, {"",}, 
-      {"volatile",  TYPE_QUAL, RID_VOLATILE},
-      {"",}, 
-      {"const",  TYPE_QUAL, RID_CONST},
-      {"",}, 
-      {"unsigned",  TYPESPEC, RID_UNSIGNED},
-      {"",}, {"",}, {"",}, {"",}, 
-      {"continue",  CONTINUE, NORID},
-      {"",}, 
-      {"register",  SCSPEC, RID_REGISTER},
-      {"",}, {"",}, {"",}, {"",}, 
-      {"goto",  GOTO, NORID},
-      {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, 
-      {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, 
-      
-      {"union",  UNION, NORID},
-      {"",}, {"",}, {"",}, {"",}, 
-      {"long",  TYPESPEC, RID_LONG},
-    };
-
   if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
     {
       register int key = hash (str, len);
@@ -256,6 +293,35 @@ make_pointer_declarator (type_quals, target)
 }
 
 void
+forget_protocol_qualifiers ()
+{
+  int i, n = sizeof wordlist / sizeof (struct resword);
+
+  for (i = 0; i < n; i++)
+    if ((int) wordlist[i].rid >= (int) RID_IN
+        && (int) wordlist[i].rid <= (int) RID_ONEWAY)
+      wordlist[i].name = "";
+}
+
+void
+remember_protocol_qualifiers ()
+{
+  int i, n = sizeof wordlist / sizeof (struct resword);
+
+  for (i = 0; i < n; i++)
+    if (wordlist[i].rid == RID_IN)
+      wordlist[i].name = "in";
+    else if (wordlist[i].rid == RID_OUT)
+      wordlist[i].name = "out";
+    else if (wordlist[i].rid == RID_INOUT)
+      wordlist[i].name = "inout";
+    else if (wordlist[i].rid == RID_BYCOPY)
+      wordlist[i].name = "bycopy";
+    else if (wordlist[i].rid == RID_ONEWAY)
+      wordlist[i].name = "oneway";   
+}
+
+void
 init_lex ()
 {
   /* Make identifier nodes long enough for the language-specific slots.  */
@@ -290,6 +356,15 @@ init_lex ()
   ridpointers[(int) RID_EXTERN] = get_identifier ("extern");
   ridpointers[(int) RID_TYPEDEF] = get_identifier ("typedef");
   ridpointers[(int) RID_REGISTER] = get_identifier ("register");
+  ridpointers[(int) RID_ITERATOR] = get_identifier ("iterator");
+  ridpointers[(int) RID_COMPLEX] = get_identifier ("complex");
+  ridpointers[(int) RID_ID] = get_identifier ("id");
+  ridpointers[(int) RID_IN] = get_identifier ("in");
+  ridpointers[(int) RID_OUT] = get_identifier ("out");
+  ridpointers[(int) RID_INOUT] = get_identifier ("inout");
+  ridpointers[(int) RID_BYCOPY] = get_identifier ("bycopy");
+  ridpointers[(int) RID_ONEWAY] = get_identifier ("oneway");
+  forget_protocol_qualifiers();
 
   /* Some options inhibit certain reserved words.
      Clear those words out of the hash table so they won't be recognized.  */
@@ -297,6 +372,9 @@ init_lex ()
   do { struct resword *s = is_reserved_word (STRING, sizeof (STRING) - 1); \
        if (s) s->name = ""; } while (0)
 
+  if (! doing_objc_thang)
+    UNSET_RESERVED_WORD ("id");
+
   if (flag_traditional)
     {
       UNSET_RESERVED_WORD ("const");
@@ -304,12 +382,16 @@ init_lex ()
       UNSET_RESERVED_WORD ("typeof");
       UNSET_RESERVED_WORD ("signed");
       UNSET_RESERVED_WORD ("inline");
+      UNSET_RESERVED_WORD ("iterator");
+      UNSET_RESERVED_WORD ("complex");
     }
   if (flag_no_asm)
     {
       UNSET_RESERVED_WORD ("asm");
       UNSET_RESERVED_WORD ("typeof");
       UNSET_RESERVED_WORD ("inline");
+      UNSET_RESERVED_WORD ("iterator");
+      UNSET_RESERVED_WORD ("complex");
     }
 }
 
@@ -331,6 +413,7 @@ yyprint (file, yychar, yylval)
     {
     case IDENTIFIER:
     case TYPENAME:
+    case OBJECTNAME:
       t = yylval.ttype;
       if (IDENTIFIER_POINTER (t))
 	fprintf (file, " `%s'", IDENTIFIER_POINTER (t));
@@ -659,7 +742,13 @@ linenum:
       while (c == ' ' || c == '\t')
 	c = getc (finput);
       if (c == '\n')
-	return c;
+	{
+	  /* Update the name in the top element of input_file_stack.  */
+	  if (input_file_stack)
+	    input_file_stack->name = input_filename;
+
+	  return c;
+	}
       ungetc (c, finput);
 
       token = yylex ();
@@ -711,6 +800,11 @@ linenum:
 	    }
 	}
 
+      /* Now that we've pushed or popped the input stack,
+	 update the name in the top element.  */
+      if (input_file_stack)
+	input_file_stack->name = input_filename;
+
       /* If we have handled a `1' or a `2',
 	 see if there is another number to read.  */
       if (used_up)
@@ -918,6 +1012,8 @@ readescape (ignore_ptr)
     case '(':
     case '{':
     case '[':
+      /* `\%' is used to prevent SCCS from getting confused.  */
+    case '%':
       if (pedantic)
 	pedwarn ("non-ANSI escape sequence `\\%c'", c);
       return c;
@@ -984,6 +1080,7 @@ yylex ()
   register char *p;
   register int value;
   int wide_flag = 0;
+  int objc_flag = 0;
 
   if (nextchar >= 0)
     c = nextchar, nextchar = -1;
@@ -1057,28 +1154,19 @@ yylex ()
 	  value = c;
 	  break;
 	}
-      p = token_buffer;
-      *p++ = '@';
-      c = getc (finput);
-      while (isalnum (c) || c == '_')
+      else
 	{
-	  if (p >= token_buffer + maxtoken)
-	    p = extend_token_buffer (p);
-
-	  *p++ = c;
-	  c = getc (finput);
+	  /* '@' may start a constant string object.  */
+	  register int c = getc(finput);
+	  if (c == '"')
+	    {
+	      objc_flag = 1;
+	      goto string_constant;
+	    }
+	  ungetc(c, finput);
+	  /* Fall through to treat '@' as the start of an indentifier.  */
 	}
 
-      *p = 0;
-      nextchar = c;
-      value = recognize_objc_keyword (token_buffer + 1);
-      if (value != 0)
-	break;
-      error ("invalid Objective C keyword `%s'", token_buffer);
-      /* Cause a syntax error--1 is not a valid token type.  */
-      value = 1;
-      break;
-
     case 'A':  case 'B':  case 'C':  case 'D':  case 'E':
     case 'F':  case 'G':  case 'H':  case 'I':  case 'J':
     case 'K':		  case 'M':  case 'N':  case 'O':
@@ -1096,11 +1184,15 @@ yylex ()
       p = token_buffer;
       while (isalnum (c) || c == '_' || c == '$' || c == '@')
 	{
-	  if (p >= token_buffer + maxtoken)
-	    p = extend_token_buffer (p);
+	  /* Make sure this char really belongs in an identifier.  */
+	  if (c == '@' && ! doing_objc_thang)
+	    break;
 	  if (c == '$' && ! dollars_in_ident)
 	    break;
 
+	  if (p >= token_buffer + maxtoken)
+	    p = extend_token_buffer (p);
+
 	  *p++ = c;
 	  c = getc (finput);
 	}
@@ -1122,6 +1214,16 @@ yylex ()
 	      yylval.ttype = ridpointers[(int) ptr->rid];
 	    value = (int) ptr->token;
 
+	    /* Only return OBJECTNAME if it is a typedef.  */
+	    if (doing_objc_thang && value == OBJECTNAME)
+	      {
+		lastiddecl = lookup_name(yylval.ttype);
+
+		if (lastiddecl == NULL_TREE
+		    || TREE_CODE (lastiddecl) != TYPE_DECL)
+		  value = IDENTIFIER;
+	      }
+
 	    /* Even if we decided to recognize asm, still perhaps warn.  */
 	    if (pedantic
 		&& (value == ASM_KEYWORD || value == TYPEOF
@@ -1137,6 +1239,9 @@ yylex ()
 
       if (value == IDENTIFIER)
 	{
+ 	  if (token_buffer[0] == '@')
+	    error("invalid identifier `%s'", token_buffer);
+
           yylval.ttype = get_identifier (token_buffer);
 	  lastiddecl = lookup_name (yylval.ttype);
 
@@ -1151,12 +1256,17 @@ yylex ()
 		   && DECL_INITIAL (lastiddecl) != 0
 		   && TREE_CODE (DECL_INITIAL (lastiddecl)) == STRING_CST)
 	    {
-	      yylval.ttype = DECL_INITIAL (lastiddecl);
+	      tree stringval = DECL_INITIAL (lastiddecl);
+	      
+	      /* Copy the string value so that we won't clobber anything
+		 if we put something in the TREE_CHAIN of this one.  */
+	      yylval.ttype = build_string (TREE_STRING_LENGTH (stringval),
+					   TREE_STRING_POINTER (stringval));
 	      value = STRING;
 	    }
           else if (doing_objc_thang)
             {
-	      tree objc_interface_decl = lookup_interface (yylval.ttype);
+	      tree objc_interface_decl = is_class_name (yylval.ttype);
 
 	      if (objc_interface_decl)
 		{
@@ -1216,8 +1326,9 @@ yylex ()
 	/* Read all the digits-and-decimal-points.  */
 
 	while (c == '.'
-	       || (isalnum (c) && (c != 'l') && (c != 'L')
-		   && (c != 'u') && (c != 'U')
+	       || (isalnum (c) && c != 'l' && c != 'L'
+		   && c != 'u' && c != 'U'
+		   && c != 'i' && c != 'I' && c != 'j' && c != 'J'
 		   && (floatflag == NOT_FLOAT || ((c != 'f') && (c != 'F')))))
 	  {
 	    if (c == '.')
@@ -1268,7 +1379,7 @@ yylex ()
 		  }
 		else if (base <= 10)
 		  {
-		    if ((c&~040) == 'E')
+		    if (c == 'e' || c == 'E')
 		      {
 			base = 10;
 			floatflag = AFTER_POINT;
@@ -1327,6 +1438,7 @@ yylex ()
 	  {
 	    tree type = double_type_node;
 	    int garbage_chars = 0, exceeds_double = 0;
+	    int imag = 0;
 	    REAL_VALUE_TYPE value;
 	    jmp_buf handler;
 
@@ -1366,41 +1478,65 @@ yylex ()
 	    else
 	      {
 		set_float_handler (handler);
-		value = REAL_VALUE_ATOF (token_buffer);
-		set_float_handler (NULL_PTR);
-	      }
-#ifdef ERANGE
-	    if (errno == ERANGE && !flag_traditional && pedantic)
-	      {
-  		/* ERANGE is also reported for underflow,
-  		   so test the value to distinguish overflow from that.  */
-		if (REAL_VALUES_LESS (dconst1, value)
-		    || REAL_VALUES_LESS (value, dconstm1))
-		  {
-		    pedwarn ("floating point number exceeds range of `double'");
-		    exceeds_double = 1;
-		  }
-	      }
-#endif
+
+/* The second argument, machine_mode, of REAL_VALUE_ATOF tells the
+   desired precision of the binary result of decimal-to-binary conversion.  */
 
 	    /* Read the suffixes to choose a data type.  */
 	    switch (c)
 	      {
 	      case 'f': case 'F':
 		type = float_type_node;
-		value = REAL_VALUE_TRUNCATE (TYPE_MODE (type), value);
-		if (REAL_VALUE_ISINF (value) && ! exceeds_double && pedantic)
+		value = REAL_VALUE_ATOF (token_buffer, TYPE_MODE (type));
+		if (TARGET_FLOAT_FORMAT != IEEE_FLOAT_FORMAT
+		    && REAL_VALUE_ISINF (value) && pedantic)
 		  pedwarn ("floating point number exceeds range of `float'");
 		garbage_chars = -1;
 		break;
 
 	      case 'l': case 'L':
 		type = long_double_type_node;
+		value = REAL_VALUE_ATOF (token_buffer, TYPE_MODE (type));
+		if (TARGET_FLOAT_FORMAT != IEEE_FLOAT_FORMAT
+		    && REAL_VALUE_ISINF (value) && pedantic)
+		  pedwarn (
+		      "floating point number exceeds range of `long double'");
 		garbage_chars = -1;
 		break;
+
+	      case 'i': case 'I':
+		if (imag)
+		  error ("more than one `i' or `j' in numeric constant");
+		imag = 1;
+		garbage_chars = -1;
+		break;
+
+              default:
+		value = REAL_VALUE_ATOF (token_buffer, TYPE_MODE (type));
+		if (TARGET_FLOAT_FORMAT != IEEE_FLOAT_FORMAT
+		    && REAL_VALUE_ISINF (value) && pedantic)
+		  pedwarn ("floating point number exceeds range of `double'");
 	      }
+	    set_float_handler (NULL_PTR);
+	    }
+#ifdef ERANGE
+	    if (errno == ERANGE && !flag_traditional && pedantic)
+	      {
+  		/* ERANGE is also reported for underflow,
+  		   so test the value to distinguish overflow from that.  */
+		if (TARGET_FLOAT_FORMAT != IEEE_FLOAT_FORMAT
+		    && (REAL_VALUES_LESS (dconst1, value)
+			|| REAL_VALUES_LESS (value, dconstm1)))
+		  {
+		    pedwarn ("floating point number exceeds range of `double'");
+		    exceeds_double = 1;
+		  }
+	      }
+#endif
 	    /* Note: garbage_chars is -1 if first char is *not* garbage.  */
-	    while (isalnum (c))
+	    while (isalnum (c) || c == '.' || c == '_'
+		   || (!flag_traditional && (c == '+' || c == '-')
+		       && (p[-1] == 'e' || p[-1] == 'E')))
 	      {
 		if (p >= token_buffer + maxtoken - 3)
 		  p = extend_token_buffer (p);
@@ -1412,7 +1548,11 @@ yylex ()
 	      error ("garbage at end of number");
 
 	    /* Create a node with determined type and value.  */
-	    yylval.ttype = build_real (type, value);
+	    if (imag)
+	      yylval.ttype = build_complex (convert (type, integer_zero_node),
+					    build_real (type, value));
+	    else
+	      yylval.ttype = build_real (type, value);
 
 	    ungetc (c, finput);
 	    *p = 0;
@@ -1424,6 +1564,7 @@ yylex ()
 	    int spec_unsigned = 0;
 	    int spec_long = 0;
 	    int spec_long_long = 0;
+	    int spec_imag = 0;
 	    int bytes, warn, i;
 
 	    while (1)
@@ -1446,12 +1587,22 @@ yylex ()
 		      }
 		    spec_long = 1;
 		  }
+		else if (c == 'i' || c == 'j' || c == 'I' || c == 'J')
+		  {
+		    if (spec_imag)
+		      error ("more than one `i' or `j' in numeric constant");
+		    spec_imag = 1;
+		  }
 		else
 		  {
-		    if (isalnum (c))
+		    if (isalnum (c) || c == '.' || c == '_'
+			|| (!flag_traditional && (c == '+' || c == '-')
+			    && (p[-1] == 'e' || p[-1] == 'E')))
 		      {
 			error ("garbage at end of number");
-			while (isalnum (c))
+			while (isalnum (c) || c == '.' || c == '_'
+			       || (!flag_traditional && (c == '+' || c == '-')
+				   && (p[-1] == 'e' || p[-1] == 'E')))
 			  {
 			    if (p >= token_buffer + maxtoken - 3)
 			      p = extend_token_buffer (p);
@@ -1580,9 +1731,19 @@ yylex ()
 	      pedwarn ("integer constant out of range");
 
 	    if (base == 10 && ! spec_unsigned && TREE_UNSIGNED (type))
-	      warning ("integer constant is so large that it is unsigned");
+	      warning ("decimal constant is so large that it is unsigned");
 
-	    if (flag_traditional && !int_fits_type_p (yylval.ttype, type))
+	    if (spec_imag)
+	      {
+		if (TYPE_PRECISION (type)
+		    <= TYPE_PRECISION (integer_type_node))
+		  yylval.ttype
+		    = build_complex (integer_zero_node,
+				     convert (integer_type_node, yylval.ttype));
+		else
+		  error ("complex integer constant is too wide for `complex int'");
+	      }
+	    else if (flag_traditional && !int_fits_type_p (yylval.ttype, type))
 	      /* The traditional constant 0x80000000 is signed
 		 but doesn't fit in the range of int.
 		 This will change it to -0x80000000, which does fit.  */
@@ -1637,6 +1798,10 @@ yylex ()
 		if (width < HOST_BITS_PER_INT
 		    && (unsigned) c >= (1 << width))
 		  pedwarn ("escape sequence out of range for character");
+#ifdef MAP_CHARACTER
+		if (isprint (c))
+		  c = MAP_CHARACTER (c);
+#endif
 	      }
 	    else if (c == '\n')
 	      {
@@ -1644,6 +1809,10 @@ yylex ()
 		  pedwarn ("ANSI C forbids newline in character constant");
 		lineno++;
 	      }
+#ifdef MAP_CHARACTER
+	    else
+	      c = MAP_CHARACTER (c);
+#endif
 
 	    num_chars++;
 	    if (num_chars > maxtoken - 4)
@@ -1680,8 +1849,11 @@ yylex ()
 	if (! wide_flag)
 	  {
 	    int num_bits = num_chars * width;
-	    if (TREE_UNSIGNED (char_type_node)
-		|| ((result >> (num_bits - 1)) & 1) == 0)
+	    if (num_bits == 0)
+	      /* We already got an error; avoid invalid shift.  */
+	      yylval.ttype = build_int_2 (0, 0);
+	    else if (TREE_UNSIGNED (char_type_node)
+		     || ((result >> (num_bits - 1)) & 1) == 0)
 	      yylval.ttype
 		= build_int_2 (result & ((unsigned HOST_WIDE_INT) ~0
 					 >> (HOST_BITS_PER_WIDE_INT - num_bits)),
@@ -1691,6 +1863,7 @@ yylex ()
 		= build_int_2 (result | ~((unsigned HOST_WIDE_INT) ~0
 					  >> (HOST_BITS_PER_WIDE_INT - num_bits)),
 			       -1);
+	    TREE_TYPE (yylval.ttype) = integer_type_node;
 	  }
 	else
 	  {
@@ -1711,9 +1884,9 @@ yylex ()
 	      }
 #endif
 	    yylval.ttype = build_int_2 (result, 0);
+	    TREE_TYPE (yylval.ttype) = wchar_type_node;
 	  }
 
-	TREE_TYPE (yylval.ttype) = integer_type_node;
 	value = CONSTANT;
 	break;
       }
@@ -1766,7 +1939,7 @@ yylex ()
 
 #ifdef MULTIBYTE_CHARS
 	    len = mbstowcs ((wchar_t *) widep, token_buffer + 1, p - token_buffer);
-	    if ((unsigned) len >= (p - token_buffer))
+	    if (len < 0 || len >= (p - token_buffer))
 	      {
 		warning ("Ignoring invalid multibyte string");
 		len = 0;
@@ -1791,17 +1964,28 @@ yylex ()
 #endif
 	    yylval.ttype = build_string ((len + 1) * WCHAR_BYTES, widep);
 	    TREE_TYPE (yylval.ttype) = wchar_array_type_node;
+	    value = STRING;
+	  }
+	else if (objc_flag)
+	  {
+	    extern tree build_objc_string();
+	    /* Return an Objective-C @"..." constant string object.  */
+	    yylval.ttype = build_objc_string (p - token_buffer,
+					      token_buffer + 1);
+	    TREE_TYPE (yylval.ttype) = char_array_type_node;
+	    value = OBJC_STRING;
 	  }
 	else
 	  {
 	    yylval.ttype = build_string (p - token_buffer, token_buffer + 1);
 	    TREE_TYPE (yylval.ttype) = char_array_type_node;
+	    value = STRING;
 	  }
 
 	*p++ = '"';
 	*p = 0;
 
-	value = STRING; break;
+	break;
       }
 
     case '+':
