@@ -1314,13 +1314,13 @@ yylex ()
 {
   register int c;
   register int namelen;
-  register char *tokstart;
+  register unsigned char *tokstart;
   register struct token *toktab;
   int wide_flag;
 
  retry:
 
-  tokstart = lexptr;
+  tokstart = (unsigned char *) lexptr;
   c = *tokstart;
   /* See if it is a special token of length 2.  */
   if (! keyword_parsing)
@@ -1376,7 +1376,7 @@ yylex ()
 	else if (c == '\'')
 	  break;
       }
-      yylval.name.address = (U_CHAR *) tokstart;
+      yylval.name.address = tokstart;
       yylval.name.length = lexptr - start_ptr;
       return NAME;
     }
@@ -1531,7 +1531,7 @@ yylex ()
 	else if (c == '"')
 	  break;
       }
-      yylval.name.address = (U_CHAR *) tokstart;
+      yylval.name.address = tokstart;
       yylval.name.length = lexptr - start_ptr;
       return NAME;
     }
@@ -1570,7 +1570,7 @@ yylex ()
   }
   
   lexptr += namelen;
-  yylval.name.address = (U_CHAR *) tokstart;
+  yylval.name.address = tokstart;
   yylval.name.length = namelen;
   return NAME;
 }
@@ -1602,6 +1602,9 @@ parse_escape (string_ptr)
     case 'b':
       return TARGET_BS;
     case 'e':
+    case 'E':
+      if (pedantic)
+	pedwarn ("non-ANSI-standard escape sequence, `\\%c'", c);
       return 033;
     case 'f':
       return TARGET_FF;
@@ -1618,13 +1621,6 @@ parse_escape (string_ptr)
     case 0:
       (*string_ptr)--;
       return 0;
-    case '^':
-      c = *(*string_ptr)++;
-      if (c == '\\')
-	c = parse_escape (string_ptr);
-      if (c == '?')
-	return 0177;
-      return (c & 0200) | (c & 037);
       
     case '0':
     case '1':
