@@ -1,5 +1,5 @@
 /* Functions related to invoking methods and overloaded functions.
-   Copyright (C) 1987, 1992 Free Software Foundation, Inc.
+   Copyright (C) 1987, 1992, 1993 Free Software Foundation, Inc.
    Contributed by Michael Tiemann (tiemann@cygnus.com)
 
 This file is part of GNU CC.
@@ -26,7 +26,6 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include <stdio.h>
 #include "cp-tree.h"
 #include "flags.h"
-#include "cp-class.h"
 
 #include "obstack.h"
 #define obstack_chunk_alloc xmalloc
@@ -34,8 +33,6 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 extern void sorry ();
 
-extern tree build_function_call_maybe ();
-
 extern int inhibit_warnings;
 extern int flag_assume_nonnull_objects;
 extern tree ctor_label, dtor_label;
@@ -48,15 +45,16 @@ extern tree unary_complex_lvalue ();
 static int convert_harshness ();
 
 #define EVIL_HARSHNESS(ARG) ((ARG) & 1)
-#define USER_HARSHNESS(ARG) ((ARG) & 2)
-#define CONTRAVARIANT_HARSHNESS(ARG) ((ARG) & 4)
-#define BASE_DERIVED_HARSHNESS(ARG) ((ARG) & 8)
-#define INT_TO_BD_HARSHNESS(ARG) (((ARG) << 4) | 8)
-#define INT_FROM_BD_HARSHNESS(ARG) ((ARG) >> 4)
-#define INT_TO_EASY_HARSHNESS(ARG) ((ARG) << 4)
-#define INT_FROM_EASY_HARSHNESS(ARG) ((ARG) >> 4)
-#define ONLY_EASY_HARSHNESS(ARG) (((ARG) & 15) == 0)
-#define CONST_HARSHNESS(ARG) ((ARG) & 1024)
+#define ELLIPSIS_HARSHNESS(ARG) ((ARG) & 2)
+#define USER_HARSHNESS(ARG) ((ARG) & 4)
+#define CONTRAVARIANT_HARSHNESS(ARG) ((ARG) & 8)
+#define BASE_DERIVED_HARSHNESS(ARG) ((ARG) & 16)
+#define INT_TO_BD_HARSHNESS(ARG) (((ARG) << 5) | 16)
+#define INT_FROM_BD_HARSHNESS(ARG) ((ARG) >> 5)
+#define INT_TO_EASY_HARSHNESS(ARG) ((ARG) << 5)
+#define INT_FROM_EASY_HARSHNESS(ARG) ((ARG) >> 5)
+#define ONLY_EASY_HARSHNESS(ARG) (((ARG) & 31) == 0)
+#define CONST_HARSHNESS(ARG) ((ARG) & 2048)
 
 /* Ordering function for overload resolution.  */
 int
@@ -67,6 +65,8 @@ rank_for_overload (x, y)
     return y->evil - x->evil;
   if (CONST_HARSHNESS (y->harshness[0]) ^ CONST_HARSHNESS (x->harshness[0]))
     return y->harshness[0] - x->harshness[0];
+  if (y->ellipsis - x->ellipsis)
+    return y->ellipsis - x->ellipsis;
   if (y->user - x->user)
     return y->user - x->user;
   if (y->b_or_d - x->b_or_d)
@@ -117,7 +117,7 @@ convert_harshness (type, parmtype, parm)
       /* Compare return types.  */
       p1 = TREE_TYPE (type);
       p2 = TREE_TYPE (parmtype);
-      new_harshness = convert_harshness (p1, p2, 0);
+      new_harshness = convert_harshness (p1, p2, NULL_TREE);
       if (new_harshness & 1)
 	return 1;
 
@@ -139,7 +139,7 @@ convert_harshness (type, parmtype, parm)
 
 	  if (! BINFO_OFFSET_ZEROP (binfo))
 	    {
-	      static int explained;
+	      static int explained = 0;
 	      if (CONTRAVARIANT_HARSHNESS (new_harshness))
 		message_2_types (sorry, "cannot cast `%d' to `%d' at function call site", p2, p1);
 	      else
@@ -155,9 +155,10 @@ convert_harshness (type, parmtype, parm)
 
       p1 = TYPE_ARG_TYPES (type);
       p2 = TYPE_ARG_TYPES (parmtype);
-      while (p1 && p2)
+      while (p1 && TREE_VALUE (p1) != void_type_node
+	     && p2 && TREE_VALUE (p2) != void_type_node)
 	{
-	  new_harshness = convert_harshness (TREE_VALUE (p1), TREE_VALUE (p2), 0);
+	  new_harshness = convert_harshness (TREE_VALUE (p1), TREE_VALUE (p2), NULL_TREE);
 	  if (EVIL_HARSHNESS (new_harshness))
 	    return 1;
 
@@ -183,12 +184,14 @@ convert_harshness (type, parmtype, parm)
       if (p1 == p2)
 	return harshness;
       if (p2)
-	return 1;
+	return p1 ? 1 : (harshness | ELLIPSIS_HARSHNESS (-1));
       if (p1)
 	return harshness | (TREE_PURPOSE (p1) == NULL_TREE);
     }
   else if (codel == POINTER_TYPE && coder == OFFSET_TYPE)
     {
+      /* XXX: Note this is set a few times, but it's never actually
+	 used! (bpk) */
       int harshness;
 
       /* Get to the OFFSET_TYPE that this might be.  */
@@ -354,7 +357,7 @@ convert_harshness (type, parmtype, parm)
 	    /* When passing a non-const argument into a const reference,
 	       dig it a little, so a non-const reference is preferred over
 	       this one. (mrs) */
-	    if (TREE_READONLY (ttl) && ! TREE_READONLY (parm))
+	    if (parm && TREE_READONLY (ttl) && ! TREE_READONLY (parm))
 	      penalty = 2;
 	    else
 	      penalty = 0;
@@ -379,17 +382,24 @@ convert_harshness (type, parmtype, parm)
 	      {
 		/* Can reference be built up?  */
 		if (ttl == intype && penalty == 0) {
-		  /* Becuase the READONLY bits are not always in the type,
-		     this extra check is necessary.  The problem should
-		     be fixed someplace else, and this extra code removed.
+		  /* Because the READONLY bits and VOLATILE bits are not
+		     always in the type, this extra check is necessary.  The
+		     problem should be fixed someplace else, and this extra
+		     code removed.
 
 		     Also, if type if a reference, the readonly bits could
 		     either be in the outer type (with reference) or on the
 		     inner type (the thing being referenced).  (mrs)  */
-		  if (TREE_READONLY (parm)
-		      && ! (TYPE_READONLY (type)
-			    || (TREE_CODE (type) == REFERENCE_TYPE
-				&& TYPE_READONLY (TREE_TYPE (type)))))
+		  if (parm
+		      && ((TREE_READONLY (parm)
+			  && ! (TYPE_READONLY (type)
+				|| (TREE_CODE (type) == REFERENCE_TYPE
+				    && TYPE_READONLY (TREE_TYPE (type)))))
+			  || (TREE_SIDE_EFFECTS (parm)
+			      && ! (TYPE_VOLATILE (type)
+				    || (TREE_CODE (type) == REFERENCE_TYPE
+					&& TYPE_VOLATILE (TREE_TYPE (type)))))))
+		     
 		    penalty = 2;
 		  else
 		    return 0;
@@ -439,7 +449,7 @@ convert_harshness (type, parmtype, parm)
 		     || coder == ENUMERAL_TYPE
 		     || coder == REAL_TYPE)
 	    && ! lvalue_p (parm))
-	  return (convert_harshness (ttl, ttr, 0)
+	  return (convert_harshness (ttl, ttr, NULL_TREE)
 		  | INT_TO_EASY_HARSHNESS (penalty));
 
 	if (ttl == ttr)
@@ -457,7 +467,7 @@ convert_harshness (type, parmtype, parm)
 	    return INT_TO_EASY_HARSHNESS (penalty+1);
 
 	if (parm && codel != REFERENCE_TYPE)
-	  return (convert_harshness (ttl, ttr, 0)
+	  return (convert_harshness (ttl, ttr, NULL_TREE)
 		  | INT_TO_EASY_HARSHNESS (penalty));
 
 	/* Here it does matter.  If this conversion is from
@@ -530,6 +540,7 @@ compute_conversion_costs (function, tta_in, cp, arglen)
 
   /* Start out with no strikes against.  */
   int evil_strikes = 0;
+  int ellipsis_strikes = 0;
   int user_strikes = 0;
   int b_or_d_strikes = 0;
   int easy_strikes = 0;
@@ -589,6 +600,10 @@ compute_conversion_costs (function, tta_in, cp, arglen)
 	  cp->u.bad_arg = strike_index;
 	  evil_strikes = 1;
 	}
+     else if (ELLIPSIS_HARSHNESS (harshness))
+	{
+	  ellipsis_strikes += 1;
+	}
 #if 0
       /* This is never set by `convert_harshness'.  */
       else if (USER_HARSHNESS (harshness))
@@ -771,6 +786,7 @@ compute_conversion_costs (function, tta_in, cp, arglen)
     }
 
   cp->evil = evil_strikes;
+  cp->ellipsis = ellipsis_strikes;
   cp->user = user_strikes;
   cp->b_or_d = b_or_d_strikes;
   cp->easy = easy_strikes;
@@ -858,7 +874,11 @@ ideal_candidate (basetype, candidates, n_candidates, parms, len)
 	 we have an ambiguity.  */
       int best = 0;
       tree tta = parms;
-      tree f1, p1;
+      tree f1;
+#if 0
+      /* for LUCID */
+      tree p1;
+#endif
 
       /* Stash all of our parameters in safe places
 	 so that we can perform type conversions in place.  */
@@ -912,7 +932,10 @@ ideal_candidate (basetype, candidates, n_candidates, parms, len)
 		{
 		  best = i;
 		  f1 = cp[best].function;
+#if 0
+		  /* For LUCID */
 		  p1 = TYPE_ARG_TYPES (TREE_TYPE (f1));
+#endif
 		}
 	      else
 		{
@@ -1042,9 +1065,9 @@ ideal_candidate (basetype, candidates, n_candidates, parms, len)
 		      && ! UNIQUELY_DERIVED_FROM_P (base1, newbase))
 		    {
 		      error_with_aggr_type (basetype, "ambiguous request for function from distinct base classes of type `%s'");
-		      error ("first candidate is `%s'",
+		      error ("  first candidate is `%s'",
 			     fndecl_as_string (0, candidates[best].function, 1));
-		      error ("second candidate is `%s'",
+		      error ("  second candidate is `%s'",
 			     fndecl_as_string (0, candidates[i].function, 1));
 		      cp[-1].evil = 1;
 		      return cp - 1;
@@ -1116,46 +1139,6 @@ may_be_remote (parent)
   return 1;
 }
 
-#ifdef ESKIT
-/* Return the number of bytes that the arglist in PARMS would
-   occupy on the stack.  */
-int
-get_arglist_len_in_bytes (parms)
-     tree parms;
-{
-  register tree parm;
-  register int bytecount = 0;
-
-  for (parm = parms; parm; parm = TREE_CHAIN (parm))
-    {
-      register tree pval = TREE_VALUE (parm);
-      register int used, size;
-
-      if (TREE_CODE (pval) == ERROR_MARK)
-	continue;
-      else if (TYPE_MODE (TREE_TYPE (pval)) != BLKmode)
-	{
-	  used = size = GET_MODE_SIZE (TYPE_MODE (TREE_TYPE (pval)));
-#ifdef PUSH_ROUNDING
-	  size = PUSH_ROUNDING (size);
-#endif
-	  used = (((size + PARM_BOUNDARY / BITS_PER_UNIT - 1)
-		   / (PARM_BOUNDARY / BITS_PER_UNIT))
-		  * (PARM_BOUNDARY / BITS_PER_UNIT));
-	}
-      else
-	{
-	  register tree size = size_in_bytes (TREE_TYPE (pval));
-	  register tree used_t
-	    = round_up (size, PARM_BOUNDARY / BITS_PER_UNIT);
-	  used = TREE_INT_CST_LOW (used_t);
-	}
-      bytecount += used;
-    }
-  return bytecount;
-}
-#endif
-
 tree
 build_vfield_ref (datum, type)
      tree datum, type;
@@ -1163,6 +1146,9 @@ build_vfield_ref (datum, type)
   tree rval;
   int old_assume_nonnull_objects = flag_assume_nonnull_objects;
 
+  if (datum == error_mark_node)
+    return error_mark_node;
+
   /* Vtable references are always made from non-null objects.  */
   flag_assume_nonnull_objects = 1;
   if (TREE_CODE (TREE_TYPE (datum)) == REFERENCE_TYPE)
@@ -1210,7 +1196,7 @@ build_field_call (basetype_path, instance_ptr, name, parms, err_name)
 
 	  if (TYPE_LANG_SPECIFIC (TREE_TYPE (instance))
 	      && TYPE_OVERLOADS_CALL_EXPR (TREE_TYPE (instance)))
-	    return build_opfncall (CALL_EXPR, LOOKUP_NORMAL, instance, parms);
+	    return build_opfncall (CALL_EXPR, LOOKUP_NORMAL, instance, parms, NULL_TREE);
 
 	  if (TREE_CODE (TREE_TYPE (instance)) == POINTER_TYPE)
 	    if (TREE_CODE (TREE_TYPE (TREE_TYPE (instance))) == FUNCTION_TYPE)
@@ -1240,10 +1226,10 @@ build_field_call (basetype_path, instance_ptr, name, parms, err_name)
 	  basetype = DECL_FIELD_CONTEXT (field);
 	  instance_ptr = convert_pointer_to (basetype, instance_ptr);
 
-	  instance = build_indirect_ref (instance_ptr, 0);
+	  instance = build_indirect_ref (instance_ptr, NULL);
 	  return build_opfncall (CALL_EXPR, LOOKUP_NORMAL,
 				 build_component_ref_1 (instance, field, 0),
-				 parms);
+				 parms, NULL_TREE);
 	}
       if (TREE_CODE (ftype) == POINTER_TYPE)
 	{
@@ -1251,8 +1237,10 @@ build_field_call (basetype_path, instance_ptr, name, parms, err_name)
 	      || TREE_CODE (TREE_TYPE (ftype)) == METHOD_TYPE)
 	    {
 	      /* This is a member which is a pointer to function.  */
-	      tree ref = build_component_ref_1 (build_indirect_ref (instance_ptr, 0),
-						field, LOOKUP_COMPLAIN);
+	      tree ref
+		= build_component_ref_1 (build_indirect_ref (instance_ptr,
+							     NULL),
+					 field, LOOKUP_COMPLAIN);
 	      if (ref == error_mark_node)
 		return error_mark_node;
 	      return build_function_call (ref, parms);
@@ -1313,14 +1301,23 @@ resolve_scope_to_name (outer_type, inner_types)
      tree outer_type, inner_types;
 {
   register tree tmp;
-  tree tags, inner_name;
+  tree inner_name;
 
   if (outer_type == NULL_TREE && current_class_type != NULL_TREE)
     {
-      /* We first try to look for a nesting in our current class context.  */
-      tree rval = resolve_scope_to_name (current_class_type, inner_types);
-      if (rval != NULL_TREE)
-	return rval;
+      /* We first try to look for a nesting in our current class context,
+         then try any enclosing classes.  */
+      tree type = current_class_type;
+      
+      while (type && (TREE_CODE (type) == RECORD_TYPE
+		      || TREE_CODE (type) == UNION_TYPE))
+        {
+          tree rval = resolve_scope_to_name (type, inner_types);
+
+	  if (rval != NULL_TREE)
+	    return rval;
+	  type = DECL_CONTEXT (TYPE_NAME (type));
+	}
     }
 
   if (TREE_CODE (inner_types) == SCOPE_REF)
@@ -1331,7 +1328,7 @@ resolve_scope_to_name (outer_type, inner_types)
   else
     {
       inner_name = inner_types;
-      inner_types = 0;
+      inner_types = NULL_TREE;
     }
 
   if (outer_type == NULL_TREE)
@@ -1413,7 +1410,7 @@ build_scoped_method_call (exp, scopes, name, parms)
 	return error_mark_node;
       if (TREE_CODE (exp) == INDIRECT_REF)
 	decl = build_indirect_ref (convert_pointer_to (binfo,
-						       build_unary_op (ADDR_EXPR, exp, 0)), 0);
+						       build_unary_op (ADDR_EXPR, exp, 0)), NULL);
       else
 	decl = build_scoped_ref (exp, scopes);
 
@@ -1487,13 +1484,9 @@ build_method_call (instance, name, parms, basetype_path, flags)
   tree last;
   int pass;
   enum visibility_type visibility;
-  int rank_for_overload ();
 
   /* Range of cases for vtable optimization.  */
-  enum vtable_needs
-    {
-      not_needed, maybe_needed, unneeded, needed,
-    };
+  enum vtable_needs { not_needed, maybe_needed, unneeded, needed };
   enum vtable_needs need_vtbl = not_needed;
 
   char *err_name;
@@ -1502,27 +1495,8 @@ build_method_call (instance, name, parms, basetype_path, flags)
   tree instance_ptr = NULL_TREE;
   int all_virtual = flag_all_virtual;
   int static_call_context = 0;
-  tree saw_private = 0;
-  tree saw_protected = 0;
-#ifdef SOS
-  /* If call is a call to a constructor, then `dtbl'
-     will first be initialized with the function table pointer
-     of the appropriate type (calling "sosFindCode" as a last
-     resort), the the call to the constructor will go through there.  */
-  tree dtbl = (flags & LOOKUP_DYNAMIC) ? TREE_VALUE (parms) : NULL_TREE;
-
-  /* Flag saying whether or not `dtbl' has been inserted into the
-     parameter list.  This is needed because we cannot tell (until
-     we have a match) whether this parameter should go in or not.
-
-     If 1, then `dtbl' is living naturally.
-     If 0, then `dtbl' is not among the parms that we know about.
-     If -1, the `dtbl' was place into the parms unnaturally.
-
-     Note that we may side-effect the parameter list, but in such a way
-     that the caller of this function would never know.  */
-  int dtbl_inserted = (flags & LOOKUP_DYNAMIC);
-#endif
+  tree saw_private = NULL_TREE;
+  tree saw_protected = NULL_TREE;
 
   /* Keep track of `const' and `volatile' objects.  */
   int constp, volatilep;
@@ -1534,7 +1508,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
   if (instance == error_mark_node
       || name == error_mark_node
       || parms == error_mark_node
-      || (instance != 0 && TREE_TYPE (instance) == error_mark_node))
+      || (instance != NULL_TREE && TREE_TYPE (instance) == error_mark_node))
     return error_mark_node;
 
   /* This is the logic that magically deletes the second argument to
@@ -1587,7 +1561,8 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	instance_ptr = instance;
       else
 	instance_ptr = build_unary_op (ADDR_EXPR, instance, 0);
-      return build_delete (basetype, instance_ptr, integer_two_node,
+      return build_delete (build_pointer_type (basetype),
+			   instance_ptr, integer_two_node,
 			   LOOKUP_NORMAL|LOOKUP_DESTRUCTOR, 0, 0);
     }
 #endif
@@ -1815,12 +1790,12 @@ build_method_call (instance, name, parms, basetype_path, flags)
 		 may be called twice, or side effects on the instance
 		 variable (such as a post-increment), may happen twice.  */
 	      instance_ptr = save_expr (instance_ptr);
-	      instance = build_indirect_ref (instance_ptr, 0);
+	      instance = build_indirect_ref (instance_ptr, NULL);
 	    }
 	  else if (TREE_CODE (TREE_TYPE (instance)) == POINTER_TYPE)
 	    {
 	      /* This happens when called for operator new ().  */
-	      instance = build_indirect_ref (instance, 0);
+	      instance = build_indirect_ref (instance, NULL);
 	    }
 
 	  need_vtbl = maybe_needed;
@@ -1841,13 +1816,12 @@ build_method_call (instance, name, parms, basetype_path, flags)
       && (! strncmp (IDENTIFIER_POINTER (name), OPERATOR_METHOD_FORMAT,
 		     OPERATOR_METHOD_LENGTH)
 	  || instance_ptr == NULL_TREE
-	  || (TYPE_OVERLOADS_METHOD_CALL_EXPR (basetype) == 0
-	      /*&& TYPE_NEEDS_WRAPPER (basetype) == 0*/ )))
+	  || (TYPE_OVERLOADS_METHOD_CALL_EXPR (basetype) == 0)))
     all_virtual = 0;
 #endif
 
   last = NULL_TREE;
-  for (parmtypes = 0, parm = parms; parm; parm = TREE_CHAIN (parm))
+  for (parmtypes = NULL_TREE, parm = parms; parm; parm = TREE_CHAIN (parm))
     {
       tree t = TREE_TYPE (TREE_VALUE (parm));
       if (TREE_CODE (t) == OFFSET_TYPE)
@@ -1901,7 +1875,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	    return error_mark_node;
 	  instance_ptr = save_expr (instance_ptr);
 	  TREE_CALLS_NEW (instance_ptr) = 1;
-	  instance = build_indirect_ref (instance_ptr, 0);
+	  instance = build_indirect_ref (instance_ptr, NULL);
 
 	  /* If it's a default argument initialized from a ctor, what we get
 	     from instance_ptr will match the arglist for the FUNCTION_DECL
@@ -1933,7 +1907,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
       else
 	tmp = get_binfo (IDENTIFIER_TYPE_VALUE (name), basetype, 0);
       
-      if (tmp != 0)
+      if (tmp != NULL_TREE)
 	{
 	  name_kind = "constructor";
 	  
@@ -1951,28 +1925,6 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	      tmplist = tree_cons (NULL_TREE, integer_type_node, TREE_CHAIN (parmtypes));
 	      TREE_CHAIN (parmtypes) = tmplist;
 	    }
-	  
-#ifdef SOS
-	  if (TYPE_DYNAMIC (basetype) && dtbl_inserted == 0)
-	    {
-	      tree parm, parmtype;
-	      dtbl = get_sos_dtable (basetype);
-	      parm = tree_cons (NULL_TREE, dtbl, TREE_CHAIN (parms));
-	      parmtype = tree_cons (NULL_TREE, build_pointer_type (ptr_type_node), TREE_CHAIN (parmtypes));
-	      TREE_CHAIN (parms) = parm;
-	      TREE_CHAIN (parmtypes) = parmtype;
-	      dtbl_inserted = -1;
-	    }
-#endif
-	  /* constructors are in very specific places.  */
-#ifdef SOS
-	  if (dtbl_inserted == -1)
-	    {
-	      TREE_CHAIN (parmtypes) = TREE_CHAIN (TREE_CHAIN (parmtypes));
-	      TREE_CHAIN (parms) = TREE_CHAIN (TREE_CHAIN (parms));
-	      dtbl_inserted = 0;
-	    }
-#endif
 	  basetype = BINFO_TYPE (tmp);
 	}
       else
@@ -2002,7 +1954,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
       struct candidate *candidates;
       struct candidate *cp;
       int len;
-      unsigned best = 2;
+      unsigned best = 1;
 
       /* This increments every time we go up the type hierarchy.
 	 The idea is to prefer a function of the derived class if possible. */
@@ -2039,7 +1991,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 		  friend_parms = tree_cons (NULL_TREE, parm, TREE_CHAIN (parms));
 		}
 	      else
-		my_friendly_assert (0, 167);
+		my_friendly_abort (167);
 
 	      cp->harshness
 		= (unsigned short *)alloca ((len+1) * sizeof (short));
@@ -2171,7 +2123,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 			{
 			  if (! DECL_STATIC_FUNCTION_P (function))
 			    TREE_VALUE (parms) = cp->arg;
-			  if (best == 2)
+			  if (best == 1)
 			    goto found_and_maybe_warn;
 			}
 		      cp++;
@@ -2199,9 +2151,9 @@ build_method_call (instance, name, parms, basetype_path, flags)
 		  == LOOKUP_SPECULATIVELY)
 		return NULL_TREE;
 	      if (flags & LOOKUP_GLOBAL)
-		error ("no global or member function `%s' defined", err_name);
+		error ("no global or non-hidden member function `%s' defined", err_name);
 	      else
-		error_with_aggr_type (save_basetype, "no member function `%s::%s'", err_name);
+		error_with_aggr_type (save_basetype, "no non-hidden member function `%s::%s' defined", err_name);
 	      return error_mark_node;
 	    }
 	  continue;
@@ -2218,7 +2170,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	    {
 	      cp = ideal_candidate (save_basetype, candidates,
 				    cp - candidates, parms, len);
-	      if (cp == 0)
+	      if (cp == (struct candidate *)0)
 		{
 		  error ("ambiguous type conversion requested for %s `%s'",
 			 name_kind, err_name);
@@ -2283,13 +2235,15 @@ build_method_call (instance, name, parms, basetype_path, flags)
 		  tree seen = saw_private;
 
 		  if (saw_private)
-		    if (saw_protected)
-		      msg = "%s %%s (and the like) are private or protected";
-		    else
-		      msg = "the %s %%s is private";
+		    {
+		      if (saw_protected)
+			msg = "%s `%%s' (and the like) are private or protected";
+		      else
+			msg = "the %s `%%s' is private";
+		    }
 		  else
 		    {
-		      msg = "the %s %%s is protected";
+		      msg = "the %s `%%s' is protected";
 		      seen = saw_protected;
 		    }
 		  sprintf (buf, msg, name_kind);
@@ -2325,7 +2279,8 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	      error ("called for const object at this point in file");
 	    }
 	  /* Not good enough for a match.  */
-	  else return error_mark_node;
+	  else
+	    return error_mark_node;
 	}
       goto found_and_ok;
     }
@@ -2336,25 +2291,31 @@ build_method_call (instance, name, parms, basetype_path, flags)
   if (visibility == visibility_private)
     {
       if (flags & LOOKUP_COMPLAIN)
-	error_with_file_and_line (DECL_SOURCE_FILE (function),
-				  DECL_SOURCE_LINE (function),
-				  TREE_PRIVATE (function)
-				  ? "%s `%s' is private"
-				  : "%s `%s' is from private base class",
-				  name_kind,
-				  lang_printable_name (function));
+	{
+	  error_with_file_and_line (DECL_SOURCE_FILE (function),
+				    DECL_SOURCE_LINE (function),
+				    TREE_PRIVATE (function)
+				    ? "%s `%s' is private"
+				    : "%s `%s' is from private base class",
+				    name_kind,
+				    lang_printable_name (function));
+	  error ("within this context");
+	}
       return error_mark_node;
     }
   else if (visibility == visibility_protected)
     {
       if (flags & LOOKUP_COMPLAIN)
-	error_with_file_and_line (DECL_SOURCE_FILE (function),
-				  DECL_SOURCE_LINE (function),
-				  TREE_PROTECTED (function)
-				  ? "%s `%s' is protected"
-				  : "%s `%s' has protected visibility from this point",
-				  name_kind,
-				  lang_printable_name (function));
+	{
+	  error_with_file_and_line (DECL_SOURCE_FILE (function),
+				    DECL_SOURCE_LINE (function),
+				    TREE_PROTECTED (function)
+				    ? "%s `%s' is protected"
+				    : "%s `%s' has protected visibility from this point",
+				    name_kind,
+				    lang_printable_name (function));
+	  error ("within this context");
+	}
       return error_mark_node;
     }
   my_friendly_abort (1);
@@ -2387,15 +2348,10 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	need_vtbl = fixed_type ? unneeded : needed;
       else
 	need_vtbl = not_needed;
-
-      if (fixed_type && DECL_ABSTRACT_VIRTUAL_P (function))
-	{
-	  error_with_decl (function, "invalid call to abstract function `%s'");
-	  return error_mark_node;
-	}
     }
 
-  if (TREE_CODE (fntype) == METHOD_TYPE && static_call_context)
+  if (TREE_CODE (fntype) == METHOD_TYPE && static_call_context
+      && !DECL_CONSTRUCTOR_P (function))
     {
       /* Let's be nice to the user for now, and give reasonable
 	 default behavior.  */
@@ -2405,7 +2361,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	  if (basetype != current_class_type)
 	    {
 	      tree binfo = get_binfo (basetype, current_class_type, 1);
-	      if (binfo == 0)
+	      if (binfo == NULL_TREE)
 		{
 		  error_not_base_type (function, current_class_type);
 		  return error_mark_node;
@@ -2445,7 +2401,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	{
 	  basetype = DECL_CLASS_CONTEXT (function);
 	  instance_ptr = convert_pointer_to (basetype, instance_ptr);
-	  instance = build_indirect_ref (instance_ptr, 0);
+	  instance = build_indirect_ref (instance_ptr, NULL);
 	}
       parms = tree_cons (NULL_TREE, instance_ptr,
 			 convert_arguments (NULL_TREE, TREE_CHAIN (TYPE_ARG_TYPES (fntype)), TREE_CHAIN (parms), NULL_TREE, sub_flags));
@@ -2464,7 +2420,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	  if (TREE_CODE (instance_ptr) == COND_EXPR)
 	    {
 	      instance_ptr = save_expr (instance_ptr);
-	      instance = build_indirect_ref (instance_ptr, 0);
+	      instance = build_indirect_ref (instance_ptr, NULL);
 	    }
 	  else if (TREE_CODE (instance_ptr) == NOP_EXPR
 		   && TREE_CODE (TREE_OPERAND (instance_ptr, 0)) == ADDR_EXPR
@@ -2476,7 +2432,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	  else if (instance == NULL_TREE
 		   || TREE_CODE (instance) != INDIRECT_REF
 		   || TREE_OPERAND (instance, 0) != instance_ptr)
-	    instance = build_indirect_ref (instance_ptr, 0);
+	    instance = build_indirect_ref (instance_ptr, NULL);
 	}
       parms = tree_cons (NULL_TREE, instance_ptr,
 			 convert_arguments (NULL_TREE, TREE_CHAIN (TYPE_ARG_TYPES (fntype)), TREE_CHAIN (parms), NULL_TREE, LOOKUP_NORMAL));
@@ -2500,14 +2456,8 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	   )
     {
       tree fn_as_int;
-#ifdef ESKIT
-      register int bytecount = 0;
 
-      parms = tree_cons (NULL_TREE, build_int_2 (bytecount, 0),
-			 TREE_CHAIN (parms));
-#else
       parms = TREE_CHAIN (parms);
-#endif
 
       if (!all_virtual && TREE_CODE (function) == FUNCTION_DECL)
 	fn_as_int = build_unary_op (ADDR_EXPR, function, 0);
@@ -2532,29 +2482,6 @@ build_method_call (instance, name, parms, basetype_path, flags)
       TREE_TYPE (result) = value_type;
 #endif
 
-#ifdef ESKIT
-      {
-	int used, size;
-
-	/* Count the number of bytes of arguments to operator->(),
-	   not to the method itself.  In the tally, don't count bytes
-	   for pointer to member function or for the bytecount.  */
-	parms = TREE_OPERAND (result, 1);
-	bytecount = get_arglist_len_in_bytes (TREE_CHAIN (TREE_CHAIN (TREE_CHAIN (parms))));
-	used = size = GET_MODE_SIZE (TYPE_MODE (TREE_TYPE (TREE_VALUE (parms))));
-#ifdef PUSH_ROUNDING
-	size = PUSH_ROUNDING (size);
-#endif
-	used = (((size + PARM_BOUNDARY / BITS_PER_UNIT - 1)
-		 / (PARM_BOUNDARY / BITS_PER_UNIT))
-		* (PARM_BOUNDARY / BITS_PER_UNIT));
-	bytecount += used;
-	TREE_CHAIN (TREE_CHAIN (parms))
-	  = tree_cons (NULL_TREE, build_int_2 (bytecount, 0),
-		       TREE_CHAIN (TREE_CHAIN (TREE_CHAIN (parms))));
-      }
-#endif
-
       return result;
     }
 #endif
@@ -2564,17 +2491,10 @@ build_method_call (instance, name, parms, basetype_path, flags)
       function = build_vfn_ref (&TREE_VALUE (parms), instance, DECL_VINDEX (function));
       TREE_TYPE (function) = build_pointer_type (fntype);
     }
-#ifdef SOS
-  else if (basetype && TYPE_DYNAMIC (basetype))
-    {
-      function = build_array_ref (dtbl, DECL_DINDEX (function));
-      TREE_TYPE (function) = build_pointer_type (fntype);
-    }
-#endif
 
-   if (TREE_CODE (function) == FUNCTION_DECL)
-     GNU_xref_call (current_function_decl,
-		    IDENTIFIER_POINTER (DECL_ASSEMBLER_NAME (function)));
+  if (TREE_CODE (function) == FUNCTION_DECL)
+    GNU_xref_call (current_function_decl,
+		   IDENTIFIER_POINTER (DECL_ASSEMBLER_NAME (function)));
 
   if (TREE_CODE (function) == FUNCTION_DECL)
     {
@@ -2635,7 +2555,6 @@ build_overload_call_real (fnname, parms, complain, final_cp, buildxxx)
   int parmlength = list_length (parms);
 
   struct candidate *candidates, *cp;
-  int rank_for_overload ();
 
   if (final_cp)
     {
@@ -2827,7 +2746,7 @@ build_overload_call_real (fnname, parms, complain, final_cp, buildxxx)
 	  struct candidate *best_cp
 	    = ideal_candidate (NULL_TREE, candidates,
 			       cp - candidates, parms, parmlength);
-	  if (best_cp == 0)
+	  if (best_cp == (struct candidate *)0)
 	    {
 	      if (complain)
 		error ("call of overloaded `%s' is ambiguous", IDENTIFIER_POINTER (fnname));
