@@ -1,5 +1,5 @@
 /* Language-level data type conversion for GNU C++.
-   Copyright (C) 1987, 1988, 1992 Free Software Foundation, Inc.
+   Copyright (C) 1987, 1988, 1992, 1993 Free Software Foundation, Inc.
    Hacked by Michael Tiemann (tiemann@cygnus.com)
 
 This file is part of GNU CC.
@@ -28,11 +28,10 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include "tree.h"
 #include "flags.h"
 #include "cp-tree.h"
-#include "cp-class.h"
+#include "convert.h"
 
-#define NULL 0
-
-extern void warn_for_assignment ();
+#undef NULL
+#define NULL (char *)0
 
 /* Change of width--truncation and extension of integers or reals--
    is represented with NOP_EXPR.  Proper functioning of many things
@@ -57,21 +56,12 @@ extern void warn_for_assignment ();
 /* Subroutines of `convert'.  */
 
 static tree
-convert_to_pointer (type, expr)
+cp_convert_to_pointer (type, expr)
      tree type, expr;
 {
   register tree intype = TREE_TYPE (expr);
   register enum tree_code form = TREE_CODE (intype);
   
-  if (integer_zerop (expr))
-    {
-      if (type == TREE_TYPE (null_pointer_node))
-	return null_pointer_node;
-      expr = build_int_2 (0, 0);
-      TREE_TYPE (expr) = type;
-      return expr;
-    }
-
   if (form == POINTER_TYPE)
     {
       intype = TYPE_MAIN_VARIANT (intype);
@@ -111,15 +101,6 @@ convert_to_pointer (type, expr)
       return build1 (NOP_EXPR, type, expr);
     }
 
-  if (form == INTEGER_TYPE || form == ENUMERAL_TYPE)
-    {
-      if (type_precision (intype) == POINTER_SIZE)
-	return build1 (CONVERT_EXPR, type, expr);
-      return convert_to_pointer (type,
-				 convert (type_for_size (POINTER_SIZE, 0),
-					  expr));
-    }
-
   my_friendly_assert (form != OFFSET_TYPE, 186);
 
   if (IS_AGGR_TYPE (intype))
@@ -136,9 +117,7 @@ convert_to_pointer (type, expr)
 	}
     }
 
-  error ("cannot convert to a pointer type");
-
-  return null_pointer_node;
+  return convert_to_pointer (type, expr);
 }
 
 /* Like convert, except permit conversions to take place which
@@ -196,7 +175,7 @@ convert_to_pointer_force (type, expr)
       return build1 (NOP_EXPR, type, expr);
     }
 
-  return convert_to_pointer (type, expr);
+  return cp_convert_to_pointer (type, expr);
 }
 
 /* We are passing something to a function which requires a reference.
@@ -227,7 +206,7 @@ build_up_reference (type, arg, flags, checkconst)
       if ((flags & LOOKUP_PROTECT) && binfo == error_mark_node)
 	return error_mark_node;
       if (basetype == NULL_TREE)
-	return error_not_base_type (target_type, argtype);
+	return (tree) error_not_base_type (target_type, argtype);
       basetype = BINFO_TYPE (binfo);
     }
 
@@ -310,7 +289,7 @@ build_up_reference (type, arg, flags, checkconst)
       if (mark_addressable (TREE_OPERAND (targ, 0)) == 0)
 	return error_mark_node;
       rval = build_binary_op (PLUS_EXPR, TREE_OPERAND (targ, 0),
-			      TREE_OPERAND (targ, 1));
+			      TREE_OPERAND (targ, 1), 1);
       TREE_TYPE (rval) = type;
       if (TREE_CONSTANT (TREE_OPERAND (targ, 1))
 	  && staticp (TREE_OPERAND (targ, 0)))
@@ -328,8 +307,9 @@ build_up_reference (type, arg, flags, checkconst)
 	    goto done;
 	  }
       }
-      /* we should have farmed out member pointers above.  */
-      my_friendly_assert (0, 188);
+
+      /* We should have farmed out member pointers above.  */
+      my_friendly_abort (188);
 
     case COMPONENT_REF:
       rval = build_component_addr (targ, build_pointer_type (argtype),
@@ -392,10 +372,10 @@ build_up_reference (type, arg, flags, checkconst)
 
 	  TREE_ADDRESSABLE (targ) = 1; /* so compiler doesn't die later */
 	  put_var_into_stack (targ);
+	  break;
 #else
 	  return error_mark_node;
 #endif
-	  break;
 	}
       /* Fall through.  */
     case VAR_DECL:
@@ -517,7 +497,7 @@ build_up_reference (type, arg, flags, checkconst)
 
  done_but_maybe_warn:
   if (checkconst && TREE_READONLY (arg) && ! TYPE_READONLY (target_type))
-    readonly_warning_or_error (arg, "conversion to reference");
+    readonly_error (arg, "conversion to reference", 1);
 
  done:
   if (TYPE_USES_COMPLEX_INHERITANCE (argtype))
@@ -541,9 +521,11 @@ build_up_reference (type, arg, flags, checkconst)
 
    FNDECL, PARMNUM, and ERRTYPE are only used when checking for use of
    volatile or const references where they aren't desired.  */
+
 tree
 convert_to_reference (decl, reftype, expr, fndecl, parmnum,
 		      errtype, strict, flags)
+
      tree decl;
      tree reftype, expr;
      tree fndecl;
@@ -578,7 +560,8 @@ convert_to_reference (decl, reftype, expr, fndecl, parmnum,
 	    return rval;
 	  if (TYPE_NEEDS_DESTRUCTOR (type))
 	    {
-	      rval = convert_to_reference (NULL_TREE, reftype, rval, NULL_TREE, -1, (char *)NULL, strict, flags);
+	      rval = convert_to_reference (NULL_TREE, reftype, rval, NULL_TREE,
+-1, (char *)NULL, strict, flags);
 	    }
 	  else
 	    {
@@ -586,7 +569,7 @@ convert_to_reference (decl, reftype, expr, fndecl, parmnum,
 	      rval = build (INIT_EXPR, type, decl, rval);
 	      rval = build (COMPOUND_EXPR, reftype, rval,
 			    convert_to_reference (NULL_TREE, reftype, decl,
-						  NULL_TREE, -1, (char *)NULL,
+						  NULL_TREE, -1, (char*)NULL,
 						  strict, flags));
 	    }
 	}
@@ -606,7 +589,7 @@ convert_to_reference (decl, reftype, expr, fndecl, parmnum,
      @@ is needed or desirable.
 
      @@ The second disjunct is provided to make references behave
-     @@ as some people think they should, i.e., an interconvertability
+     @@ as some people think they should, i.e., an interconvertibility
      @@ between references to builtin types (such as short and
      @@ unsigned short).  There should be no conversion between
      @@ types whose codes are different, or whose sizes are different.  */
@@ -632,7 +615,7 @@ convert_to_reference (decl, reftype, expr, fndecl, parmnum,
 				 errtype, fndecl, parmnum, pedantic);
 	  if (! TYPE_VOLATILE (ttl) && TYPE_VOLATILE (ttr))
 	    warn_for_assignment ("%s of non-`volatile &' reference from `volatile &'",
-				 "reference to volatile given for argument %d of `%s'",
+				 "reference to volatile given for argument %d of `%s'",	
 				 errtype, fndecl, parmnum, pedantic);
 	}
 
@@ -678,7 +661,6 @@ convert_to_reference (decl, reftype, expr, fndecl, parmnum,
       tree init = build_method_call (NULL_TREE, constructor_name (type),
 				     build_tree_list (NULL_TREE, expr),
 				     TYPE_BINFO (type), LOOKUP_NO_CONVERSION);
-      tree rval1;
 
       if (init != error_mark_node)
 	if (rval)
@@ -689,7 +671,7 @@ convert_to_reference (decl, reftype, expr, fndecl, parmnum,
 
       init = build_method_call (NULL_TREE, constructor_name (type),
 				build_tree_list (NULL_TREE, expr),
-				TYPE_BINFO (type), LOOKUP_NORMAL);
+				TYPE_BINFO (type), LOOKUP_NORMAL|LOOKUP_NO_CONVERSION);
 
       if (init == error_mark_node)
 	return error_mark_node;
@@ -699,12 +681,6 @@ convert_to_reference (decl, reftype, expr, fndecl, parmnum,
       return build_up_reference (reftype, rval, flags, decl!=NULL_TREE);
     }
 
-  if (rval)
-    {
-      /* If we found a way to convert earlier, then use it. */
-      return rval;
-    }
-
   my_friendly_assert (form != OFFSET_TYPE, 189);
 
   /* This is in two pieces for now, because pointer to first becomes
@@ -728,360 +704,30 @@ convert_from_reference (val)
  if (TREE_CODE (type) == REFERENCE_TYPE)
     {
       tree target_type = TREE_TYPE (type);
+      tree nval;
 
       /* This can happen if we cast to a reference type.  */
       if (TREE_CODE (val) == ADDR_EXPR)
 	{
-	  val = build1 (NOP_EXPR, build_pointer_type (target_type), val);
-	  val = build_indirect_ref (val, 0);
-	  return val;
+	  nval = build1 (NOP_EXPR, build_pointer_type (target_type), val);
+	  nval = build_indirect_ref (nval, 0);
+	  /* The below was missing, are other important flags missing too? */
+	  TREE_SIDE_EFFECTS (nval) = TREE_SIDE_EFFECTS (val);
+	  return nval;
 	}
 
-      val = build1 (INDIRECT_REF, TYPE_MAIN_VARIANT (target_type), val);
+      nval = build1 (INDIRECT_REF, TYPE_MAIN_VARIANT (target_type), val);
 
-      TREE_THIS_VOLATILE (val) = TYPE_VOLATILE (target_type);
-      TREE_SIDE_EFFECTS (val) = TYPE_VOLATILE (target_type);
-      TREE_READONLY (val) = TYPE_READONLY (target_type);
+      TREE_THIS_VOLATILE (nval) = TYPE_VOLATILE (target_type);
+      TREE_SIDE_EFFECTS (nval) = TYPE_VOLATILE (target_type);
+      TREE_READONLY (nval) = TYPE_READONLY (target_type);
+      /* The below was missing, are other important flags missing too? */
+      TREE_SIDE_EFFECTS (nval) = TREE_SIDE_EFFECTS (val);
+      return nval;
     }
   return val;
 }
-
-static tree
-convert_to_real (type, expr)
-     tree type, expr;
-{
-  register enum tree_code form = TREE_CODE (TREE_TYPE (expr));
-
-  if (form == REAL_TYPE)
-    return build1 (flag_float_store ? CONVERT_EXPR : NOP_EXPR,
-		  type, expr);
-
-  if (form == INTEGER_TYPE || form == ENUMERAL_TYPE)
-    return build1 (FLOAT_EXPR, type, expr);
-
-  my_friendly_assert (form != OFFSET_TYPE, 190);
-
-  if (form == POINTER_TYPE)
-    error ("pointer value used where a floating point value was expected");
-  /* C++: check to see if we can convert this aggregate type
-     into the required scalar type.  */
-  else if (IS_AGGR_TYPE (TREE_TYPE (expr)))
-    {
-      tree rval;
-      rval = build_type_conversion (CONVERT_EXPR, type, expr, 1);
-      if (rval)
-	return rval;
-      else
-	error ("aggregate value used where a floating point value was expected");
-    }
-
-  {
-    register tree tem = make_node (REAL_CST);
-    TREE_TYPE (tem) = type;
-    TREE_REAL_CST (tem) = REAL_VALUE_ATOF ("0.0");
-    return tem;
-  }
-}
 
-/* The result of this is always supposed to be a newly created tree node
-   not in use in any existing structure.  */
-
-static tree
-convert_to_integer (type, expr)
-     tree type, expr;
-{
-  register tree intype = TREE_TYPE (expr);
-  register enum tree_code form = TREE_CODE (intype);
-  extern tree build_binary_op_nodefault ();
-  extern tree build_unary_op ();
-
-  if (form == POINTER_TYPE)
-    {
-      if (integer_zerop (expr))
-	expr = integer_zero_node;
-      else
-	expr = fold (build1 (CONVERT_EXPR,
-			     type_for_size (POINTER_SIZE, 0), expr));
-      intype = TREE_TYPE (expr);
-      form = TREE_CODE (intype);
-      if (intype == type)
-	return expr;
-    }
-
-  if (form == INTEGER_TYPE || form == ENUMERAL_TYPE)
-    {
-      register unsigned outprec = TYPE_PRECISION (type);
-      register unsigned inprec = TYPE_PRECISION (intype);
-      register enum tree_code ex_form = TREE_CODE (expr);
-
-      if (flag_int_enum_equivalence == 0
-	  && TREE_CODE (type) == ENUMERAL_TYPE
-	  && form == INTEGER_TYPE)
-	{
-	  if (pedantic)
-	    pedwarn ("anachronistic conversion from integer type to enumeral type `%s'",
-		     TYPE_NAME_STRING (type));
-	  if (flag_pedantic_errors)
-	    return error_mark_node;
-	}
-
-      /* If we are widening the type, put in an explicit conversion.
-	 Similarly if we are not changing the width.  However, if this is
-	 a logical operation that just returns 0 or 1, we can change the
-	 type of the expression (see below).  */
-
-      if (TREE_CODE_CLASS (ex_form) == '<'
-	  || ex_form == TRUTH_AND_EXPR || ex_form == TRUTH_ANDIF_EXPR
-	  || ex_form == TRUTH_OR_EXPR || ex_form == TRUTH_ORIF_EXPR
-	  || ex_form == TRUTH_NOT_EXPR)
-	{
-	  TREE_TYPE (expr) = type;
-	  return expr;
-	}
-      else if (outprec >= inprec)
-	return build1 (NOP_EXPR, type, expr);
-
-/* Here detect when we can distribute the truncation down past some arithmetic.
-   For example, if adding two longs and converting to an int,
-   we can equally well convert both to ints and then add.
-   For the operations handled here, such truncation distribution
-   is always safe.
-   It is desirable in these cases:
-   1) when truncating down to full-word from a larger size
-   2) when truncating takes no work.
-   3) when at least one operand of the arithmetic has been extended
-   (as by C's default conversions).  In this case we need two conversions
-   if we do the arithmetic as already requested, so we might as well
-   truncate both and then combine.  Perhaps that way we need only one.
-
-   Note that in general we cannot do the arithmetic in a type
-   shorter than the desired result of conversion, even if the operands
-   are both extended from a shorter type, because they might overflow
-   if combined in that type.  The exceptions to this--the times when
-   two narrow values can be combined in their narrow type even to
-   make a wider result--are handled by "shorten" in build_binary_op.  */
-
-      switch (ex_form)
-	{
-	case RSHIFT_EXPR:
-	  /* We can pass truncation down through right shifting
-	     when the shift count is a nonpositive constant.  */
-	  if (TREE_CODE (TREE_OPERAND (expr, 1)) == INTEGER_CST
-	      && tree_int_cst_lt (TREE_OPERAND (expr, 1), integer_one_node))
-	    goto trunc1;
-	  break;
-
-	case LSHIFT_EXPR:
-	  /* We can pass truncation down through left shifting
-	     when the shift count is a nonnegative constant.  */
-	  if (TREE_CODE (TREE_OPERAND (expr, 1)) == INTEGER_CST
-	      && ! tree_int_cst_lt (TREE_OPERAND (expr, 1), integer_zero_node)
-	      && TREE_CODE (TYPE_SIZE (type)) == INTEGER_CST)
-	    {
-	      /* If shift count is less than the width of the truncated type,
-		 really shift.  */
-	      if (tree_int_cst_lt (TREE_OPERAND (expr, 1), TYPE_SIZE (type)))
-		/* In this case, shifting is like multiplication.  */
-		goto trunc1;
-	      else
-		/* If it is >= that width, result is zero.
-		   Handling this with trunc1 would give the wrong result:
-		   (int) ((long long) a << 32) is well defined (as 0)
-		   but (int) a << 32 is undefined and would get a warning.  */
-		return convert_to_integer (type, integer_zero_node);
-	    }
-	  break;
-
-	case MAX_EXPR:
-	case MIN_EXPR:
-	case MULT_EXPR:
-	  {
-	    tree arg0 = get_unwidened (TREE_OPERAND (expr, 0), type);
-	    tree arg1 = get_unwidened (TREE_OPERAND (expr, 1), type);
-
-	    /* Don't distribute unless the output precision is at least as big
-	       as the actual inputs.  Otherwise, the comparison of the
-	       truncated values will be wrong.  */
-	    if (outprec >= TYPE_PRECISION (TREE_TYPE (arg0))
-		&& outprec >= TYPE_PRECISION (TREE_TYPE (arg1))
-		/* If signedness of arg0 and arg1 don't match,
-		   we can't necessarily find a type to compare them in.  */
-		&& (TREE_UNSIGNED (TREE_TYPE (arg0))
-		    == TREE_UNSIGNED (TREE_TYPE (arg1))))
-	      goto trunc1;
-	    break;
-	  }
-
-	case PLUS_EXPR:
-	case MINUS_EXPR:
-	case BIT_AND_EXPR:
-	case BIT_IOR_EXPR:
-	case BIT_XOR_EXPR:
-	case BIT_ANDTC_EXPR:
-	trunc1:
-	  {
-	    tree arg0 = get_unwidened (TREE_OPERAND (expr, 0), type);
-	    tree arg1 = get_unwidened (TREE_OPERAND (expr, 1), type);
-
-	    if (outprec >= BITS_PER_WORD
-		|| TRULY_NOOP_TRUNCATION (outprec, inprec)
-		|| inprec > TYPE_PRECISION (TREE_TYPE (arg0))
-		|| inprec > TYPE_PRECISION (TREE_TYPE (arg1)))
-	      {
-		/* Do the arithmetic in type TYPEX,
-		   then convert result to TYPE.  */
-		register tree typex = type;
-
-		/* Can't do arithmetic in enumeral types
-		   so use an integer type that will hold the values.  */
-		if (TREE_CODE (typex) == ENUMERAL_TYPE)
-		  typex = type_for_size (TYPE_PRECISION (typex),
-					 TREE_UNSIGNED (typex));
-
-		/* But now perhaps TYPEX is as wide as INPREC.
-		   In that case, do nothing special here.
-		   (Otherwise would recurse infinitely in convert.  */
-		if (TYPE_PRECISION (typex) != inprec)
-		  {
-		    /* Don't do unsigned arithmetic where signed was wanted,
-		       or vice versa.
-		       Exception 1: if we will eventually truncate
-		       the result, then do the work as unsigned;
-		       this can prevent unnecessary sign-extension.
-		       Exception 2: if either of the original operands were
-		       unsigned then can safely do the work as unsigned.
-		       And we may need to do it as unsigned
-		       if we truncate to the original size.  */
-		    typex = (((outprec <= TYPE_PRECISION (typex)
-			       && TREE_UNSIGNED (type))
-			      || TREE_UNSIGNED (TREE_TYPE (expr))
-			      || TREE_UNSIGNED (TREE_TYPE (arg0))
-			      || TREE_UNSIGNED (TREE_TYPE (arg1)))
-			     ? unsigned_type (typex) : signed_type (typex));
-		    return convert (type,
-				    build_binary_op_nodefault (ex_form,
-							       convert (typex, arg0),
-							       convert (typex, arg1),
-							       ex_form));
-		  }
-	      }
-	  }
-	  break;
-
-	case NEGATE_EXPR:
-	case BIT_NOT_EXPR:
-	case ABS_EXPR:
-	  {
-	    register tree typex = type;
-
-	    /* Can't do arithmetic in enumeral types
-	       so use an integer type that will hold the values.  */
-	    if (TREE_CODE (typex) == ENUMERAL_TYPE)
-	      typex = type_for_size (TYPE_PRECISION (typex),
-				     TREE_UNSIGNED (typex));
-
-	    /* But now perhaps TYPEX is as wide as INPREC.
-	       In that case, do nothing special here.
-	       (Otherwise would recurse infinitely in convert.  */
-	    if (TYPE_PRECISION (typex) != inprec)
-	      {
-		/* Don't do unsigned arithmetic where signed was wanted,
-		   or vice versa.  */
-		typex = (((outprec <= TYPE_PRECISION (typex)
-			   && TREE_UNSIGNED (type))
-			  || TREE_UNSIGNED (TREE_TYPE (expr)))
-			 ? unsigned_type (typex) : signed_type (typex));
-		return convert (type,
-				build_unary_op (ex_form,
-						convert (typex, TREE_OPERAND (expr, 0)),
-						1));
-	      }
-	  }
-
-	case NOP_EXPR:
-	  /* If truncating after truncating, might as well do all at once.
-	     If truncating after extending, we may get rid of wasted work.  */
-	  return convert (type, get_unwidened (TREE_OPERAND (expr, 0), type));
-
-	case COND_EXPR:
-	  /* Can treat the two alternative values like the operands
-	     of an arithmetic expression.  */
-	  {
-	    tree arg1 = get_unwidened (TREE_OPERAND (expr, 1), type);
-	    tree arg2 = get_unwidened (TREE_OPERAND (expr, 2), type);
-
-	    if (outprec >= BITS_PER_WORD
-		|| TRULY_NOOP_TRUNCATION (outprec, inprec)
-		|| inprec > TYPE_PRECISION (TREE_TYPE (arg1))
-		|| inprec > TYPE_PRECISION (TREE_TYPE (arg2)))
-	      {
-		/* Do the arithmetic in type TYPEX,
-		   then convert result to TYPE.  */
-		register tree typex = type;
-
-		/* Can't do arithmetic in enumeral types
-		   so use an integer type that will hold the values.  */
-		if (TREE_CODE (typex) == ENUMERAL_TYPE)
-		  typex = type_for_size (TYPE_PRECISION (typex),
-					 TREE_UNSIGNED (typex));
-
-		/* But now perhaps TYPEX is as wide as INPREC.
-		   In that case, do nothing special here.
-		   (Otherwise would recurse infinitely in convert.  */
-		if (TYPE_PRECISION (typex) != inprec)
-		  {
-		    /* Don't do unsigned arithmetic where signed was wanted,
-		       or vice versa.  */
-		    typex = (((outprec <= TYPE_PRECISION (typex)
-			       && TREE_UNSIGNED (type))
-			      || TREE_UNSIGNED (TREE_TYPE (expr)))
-			     ? unsigned_type (typex) : signed_type (typex));
-		    return convert (type,
-				    fold (build (COND_EXPR, typex,
-						 TREE_OPERAND (expr, 0),
-						 convert (typex, arg1),
-						 convert (typex, arg2))));
-		  }
-		else
-		  /* It is sometimes worthwhile
-		     to push the narrowing down through the conditional.  */
-		  return fold (build (COND_EXPR, type,
-				      TREE_OPERAND (expr, 0),
-				      convert (type, TREE_OPERAND (expr, 1)), 
-				      convert (type, TREE_OPERAND (expr, 2))));
-	      }
-	  }
-
-	}
-
-      return build1 (NOP_EXPR, type, expr);
-    }
-
-  if (form == REAL_TYPE)
-    return build1 (FIX_TRUNC_EXPR, type, expr);
-
-  if (form == OFFSET_TYPE)
-    error_with_decl (TYPE_NAME (TYPE_OFFSET_BASETYPE (intype)),
-		     "pointer-to-member expression object not composed with type `%s' object");
-  else
-    {
-      if (IS_AGGR_TYPE (intype))
-	{
-	  tree rval;
-	  rval = build_type_conversion (CONVERT_EXPR, type, expr, 1);
-	  if (rval) return rval;
-	}
-
-      error ("aggregate value used where an integer was expected");
-    }
-
-  {
-    register tree tem = build_int_2 (0, 0);
-    TREE_TYPE (tem) = type;
-    return tem;
-  }
-}
-
 /* See if there is a constructor of type TYPE which will convert
    EXPR.  The reference manual seems to suggest (8.5.6) that we need
    not worry about finding constructors for base classes, then converting
@@ -1336,12 +982,14 @@ convert_pointer_to (binfo, expr)
 
   my_friendly_assert (!integer_zerop (expr), 191);
 
-  if (IS_AGGR_TYPE (type)
-      && IS_AGGR_TYPE (TREE_TYPE (intype))
+  if (TREE_CODE (type) == RECORD_TYPE
+      && TREE_CODE (TREE_TYPE (intype)) == RECORD_TYPE
       && type != TYPE_MAIN_VARIANT (TREE_TYPE (intype)))
     {
       tree path;
-      int distance = get_base_distance (type, TYPE_MAIN_VARIANT (TREE_TYPE (intype)), 0, &path);
+      int distance
+	= get_base_distance (type, TYPE_MAIN_VARIANT (TREE_TYPE (intype)),
+			     0, &path);
 
       /* This function shouldn't be called with unqualified arguments
 	 but if it is, give them an error message that they can read.
@@ -1446,11 +1094,47 @@ convert (type, expr)
     e = convert_from_reference (e);
 
   if (code == INTEGER_TYPE || code == ENUMERAL_TYPE)
-    return fold (convert_to_integer (type, e));
+    {
+      tree intype = TREE_TYPE (expr);
+      enum tree_code form = TREE_CODE (intype);
+      if (flag_int_enum_equivalence == 0
+	  && TREE_CODE (type) == ENUMERAL_TYPE
+	  && form == INTEGER_TYPE)
+	{
+	  if (pedantic)
+	    pedwarn ("anachronistic conversion from integer type to enumeral type `%s'",
+		     TYPE_NAME_STRING (type));
+	  if (flag_pedantic_errors)
+	    return error_mark_node;
+	}
+      if (form == OFFSET_TYPE)
+	error_with_decl (TYPE_NAME (TYPE_OFFSET_BASETYPE (intype)),
+			 "pointer-to-member expression object not composed with type `%s' object");
+      else if (IS_AGGR_TYPE (intype))
+	{
+	  tree rval;
+	  rval = build_type_conversion (CONVERT_EXPR, type, expr, 1);
+	  if (rval) return rval;
+	  error ("aggregate value used where an integer was expected");
+	  return error_mark_node;
+	}
+      return fold (convert_to_integer (type, e));
+    }
   if (code == POINTER_TYPE)
-    return fold (convert_to_pointer (type, e));
+    return fold (cp_convert_to_pointer (type, e));
   if (code == REAL_TYPE)
-    return fold (convert_to_real (type, e));
+    {
+      if (IS_AGGR_TYPE (TREE_TYPE (e)))
+	{
+	  tree rval;
+	  rval = build_type_conversion (CONVERT_EXPR, type, e, 1);
+	  if (rval)
+	    return rval;
+	  else
+	    error ("aggregate value used where a floating point value was expected");
+	}
+      return fold (convert_to_real (type, e));
+    }
 
   /* New C++ semantics:  since assignment is now based on
      memberwise copying,  if the rhs type is derived from the
@@ -1780,7 +1464,6 @@ build_type_conversion (code, xtype, expr, for_sure)
     return NULL_TREE;
   else
     {
-      extern tree default_conversion ();
       tree tmp = default_conversion (build1 (NOP_EXPR, type, integer_zero_node));
       if (tmp == error_mark_node)
 	return NULL_TREE;
