@@ -1,5 +1,5 @@
 /* Language-dependent node constructors for parse phase of GNU compiler.
-   Copyright (C) 1987, 1988, 1992 Free Software Foundation, Inc.
+   Copyright (C) 1987, 1988, 1992, 1993 Free Software Foundation, Inc.
    Hacked by Michael Tiemann (tiemann@cygnus.com)
 
 This file is part of GNU CC.
@@ -282,10 +282,8 @@ extern struct obstack *saveable_obstack;
 #define TYPE_HASH(TYPE) ((HOST_WIDE_INT) (TYPE) & 0777777)
 
 /* Construct, lay out and return the type of methods belonging to class
-   BASETYPE and whose arguments and values are described by TYPE.
-   If that type exists already, reuse it.
-   TYPE must be a FUNCTION_TYPE node.  */
-
+   BASETYPE and whose arguments are described by ARGTYPES and whose values
+   are described by RETTYPE.  If each type exists already, reuse it.  */
 tree
 build_cplus_method_type (basetype, rettype, argtypes)
      tree basetype, rettype, argtypes;
@@ -612,7 +610,12 @@ layout_basetypes (rec, binfos)
 
       if (TYPE_SIZE (basetype) == 0)
 	{
-	  error_with_aggr_type (base_binfo, "base class `%s' has incomplete type");
+#if 0
+	  /* This error is now reported in xref_tag, thus giving better
+	     location information.  */
+	  error_with_aggr_type (base_binfo,
+				"base class `%s' has incomplete type");
+
 	  TREE_VIA_PUBLIC (base_binfo) = 1;
 	  TREE_VIA_PROTECTED (base_binfo) = 0;
 	  TREE_VIA_VIRTUAL (base_binfo) = 0;
@@ -624,6 +627,7 @@ layout_basetypes (rec, binfos)
 
 	     does not dump core when compiled. */
 	  my_friendly_abort (121);
+#endif
 	  continue;
 	}
 
@@ -975,9 +979,8 @@ list_hash_lookup_or_cons (value)
 	{
 	  /* If the type name and constructor name are different, don't
 	     write constructor name into type.  */
-	  extern tree constructor_name ();
-	  if (IDENTIFIER_TYPEDECL_VALUE (value)
-	      && IDENTIFIER_TYPEDECL_VALUE (value) != constructor_name (value))
+	  if (identifier_typedecl_value (value)
+	      && identifier_typedecl_value (value) != constructor_name (value))
 	    list = tree_cons (NULL_TREE, value, NULL_TREE);
 	  else
 	    {
@@ -1099,69 +1102,19 @@ copy_binfo (list)
   return rval;
 }
 
-/* Return the binfo value for ELEM in TYPE.  Due to structure
-   sharing, we may find ELEM only in the association list
-   belonging to a basetype of TYPE.
+/* Return the binfo value for ELEM in TYPE.  */
 
-   COPYING is 0 if we just want an binfo value without needing
-   to modify it.
-   COPYING is 1 if we want the binfo value in order to modify it.
-   In this case, if we don't find ELEM immediately in the binfo
-   values of TYPE, we return a copy.
-   COPYING is -1 if we are called recursively and need a copy.
-   In this case we return a copy of ELEM at the point we find it.  */
 tree
-binfo_value (elem, type, copying)
+binfo_value (elem, type)
      tree elem;
      tree type;
-     int copying;
 {
-  tree binfo = TYPE_BINFO (type);
-  tree last;
-  tree rval = NULL_TREE;
-
-  /* Dispose quickly of degenerate case.  */
+  if (get_base_distance (elem, type, 0, (tree *)0) == -2)
+    compiler_error ("base class `%s' ambiguous in binfo_value",
+		    TYPE_NAME_STRING (elem));
   if (elem == type)
-    return copying < 0 ? copy_binfo (binfo) : binfo;
-
-  /* Look for ELEM in two passes.  First pass checks the entire binfo list.
-     Second pass recursively searches the binfo lists of binfos.  */
-  while (binfo)
-    {
-      if (elem == BINFO_TYPE (binfo))
-	/* If we find it on the main spine, then
-	   there can be no ambiguity.  */
-	return copying < 0 ? copy_binfo (binfo) : binfo;
-      last = binfo;
-      binfo = TREE_CHAIN (binfo);
-    }
-
-  for (binfo = TYPE_BINFO (type);
-       binfo != TREE_CHAIN (last);
-       binfo = TREE_CHAIN (binfo))
-    {
-      /* ??? Should this condition instead test
-	 BINFO_TYPE (binfo) != TYPE_MAIN_VARIANT (type) ???  */
-      if (BINFO_TYPE (binfo) != TYPE_MAIN_VARIANT (type))
-	{
-	  tree nval = binfo_value (elem, BINFO_TYPE (binfo), copying ? -1 : 0);
-
-	  if (nval)
-	    {
-	      if (copying && rval == NULL_TREE)
-		chainon (TYPE_BINFO (type), nval);
-
-	      if (rval && BINFO_TYPE (rval) != BINFO_TYPE (nval))
-		/* If we find it underneath, we must make sure that
-		   there are no two ways to do it.  */
-		compiler_error ("base class `%s' ambiguous in binfo_value",
-				TYPE_NAME_STRING (elem));
-	      else
-		rval = nval;
-	    }
-	}
-    }
-  return rval;
+    return TYPE_BINFO (type);
+  return get_binfo (elem, type, 0);
 }
 
 tree
@@ -1198,7 +1151,7 @@ virtual_member (elem, list)
       if (binfos != NULL_TREE)
 	for (i = TREE_VEC_LENGTH (binfos)-1; i >= 0; i--)
 	  {
-	    nval = binfo_value (elem, BINFO_TYPE (TREE_VEC_ELT (binfos, i)), 0);
+	    nval = binfo_value (elem, BINFO_TYPE (TREE_VEC_ELT (binfos, i)));
 	    if (nval)
 	      {
 		if (rval && BINFO_OFFSET (nval) != BINFO_OFFSET (rval))
@@ -1239,7 +1192,7 @@ virtual_offset (elem, list, initial_offset)
 
       for (i = TREE_VEC_LENGTH (binfos)-1; i >= 0; i--)
 	{
-	  nval = binfo_value (elem, BINFO_TYPE (TREE_VEC_ELT (binfos, i)), 0);
+	  nval = binfo_value (elem, BINFO_TYPE (TREE_VEC_ELT (binfos, i)));
 	  if (nval)
 	    {
 	      if (rval && BINFO_OFFSET (nval) != BINFO_OFFSET (rval))
