@@ -1,6 +1,6 @@
 /* Report error messages, build initializers, and perform
    some front-end optimizations for C++ compiler.
-   Copyright (C) 1987, 1988, 1989, 1992 Free Software Foundation, Inc.
+   Copyright (C) 1987, 1988, 1989, 1992, 1993 Free Software Foundation, Inc.
    Hacked by Michael Tiemann (tiemann@cygnus.com)
 
 This file is part of GNU CC.
@@ -36,10 +36,7 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include "flags.h"
 
 static tree process_init_constructor ();
-tree digest_init ();
-void incomplete_type_error ();
-void readonly_warning_or_error ();
-extern tree convert_for_initialization ();
+extern void pedwarn (), error ();
 
 extern int errorcount;
 extern int sorrycount;
@@ -92,7 +89,7 @@ void
 error_with_aggr_type (type, msg, arg)
      tree type;
      char *msg;
-     int arg;
+     HOST_WIDE_INT arg;
 {
   tree name;
 
@@ -105,24 +102,23 @@ error_with_aggr_type (type, msg, arg)
   error (msg, IDENTIFIER_POINTER (name), arg);
 }
 
-/* Warn or give error about storing in something that is `const'.  */
-/* for now, because the compiler is getting pickier about const'ness
-   give users a non-fatal warning instead of an error, this will
-   revert back to a hard error after a release or two.  This is
-   to not screw the user that has valid code for which the compiler
-   does not treat it right.  */
-#if 0
-#define const_error error
-#else
-#define const_error pedwarn
-#endif
-
+/* According to ARM $7.1.6, "A `const' object may be initialized, but its
+   value may not be changed thereafter.  Thus, we emit hard errors for these,
+   rather than just pedwarns.  If `SOFT' is 1, then we just pedwarn.  (For
+   example, conversions to references.)  */
 void
-readonly_warning_or_error (arg, string)
+readonly_error (arg, string, soft)
      tree arg;
      char *string;
+     int soft;
 {
   char *fmt;
+  void (*fn)();
+
+  if (soft)
+    fn = pedwarn;
+  else
+    fn = error;
 
   if (TREE_CODE (arg) == COMPONENT_REF)
     {
@@ -130,7 +126,7 @@ readonly_warning_or_error (arg, string)
         fmt = "%s of member `%s' in read-only structure";
       else
         fmt = "%s of read-only member `%s'";
-      const_error (fmt, string, lang_printable_name (TREE_OPERAND (arg, 1)));
+      (*fn) (fmt, string, lang_printable_name (TREE_OPERAND (arg, 1)));
     }
   else if (TREE_CODE (arg) == VAR_DECL)
     {
@@ -140,25 +136,19 @@ readonly_warning_or_error (arg, string)
 	fmt = "%s of constant field `%s'";
       else
 	fmt = "%s of read-only variable `%s'";
-      const_error (fmt, string, lang_printable_name (arg));
+      (*fn) (fmt, string, lang_printable_name (arg));
     }
   else if (TREE_CODE (arg) == PARM_DECL)
-    {
-      const_error ("%s of read-only parameter `%s'",
-		   string, lang_printable_name (arg));
-    }
+    (*fn) ("%s of read-only parameter `%s'", string,
+	   lang_printable_name (arg));
   else if (TREE_CODE (arg) == INDIRECT_REF
            && TREE_CODE (TREE_TYPE (TREE_OPERAND (arg, 0))) == REFERENCE_TYPE
            && (TREE_CODE (TREE_OPERAND (arg, 0)) == VAR_DECL
                || TREE_CODE (TREE_OPERAND (arg, 0)) == PARM_DECL))
-    {
-      const_error ("%s of read-only reference `%s'",
-		   string, lang_printable_name (TREE_OPERAND (arg, 0)));
-    }
+    (*fn) ("%s of read-only reference `%s'",
+	   string, lang_printable_name (TREE_OPERAND (arg, 0)));
   else	       
-    {
-      const_error ("%s of read-only location", string);
-    }
+    (*fn) ("%s of read-only location", string);
 }
 
 /* Print an error message for invalid use of a type which declares
@@ -192,15 +182,15 @@ abstract_virtuals_error (decl, type)
   /* Only go through this once.  */
   if (TREE_PURPOSE (u) == NULL_TREE)
     {
-      error ("since the following virtual functions are abstract:");
+      error ("  since the following virtual functions are abstract:");
       TREE_PURPOSE (u) = error_mark_node;
       while (u)
 	{
-	  error_with_decl (TREE_VALUE (u), "%s");
+	  error_with_decl (TREE_VALUE (u), "\t%s");
 	  u = TREE_CHAIN (u);
 	}
     }
-  else error ("since type `%s' has abstract virtual functions", typename);
+  else error ("  since type `%s' has abstract virtual functions", typename);
 }
 
 /* Print an error message for invalid use of an incomplete type.
@@ -286,23 +276,28 @@ incomplete_type_error (value, type)
    can get away with calling abort, because for most of the calls to
    abort on most machines, it, I suspect, can be proven that it is
    impossible to ever call abort.  The same is not yet true for C++,
-   one day, maybe it will be.  (mrs) */
+   one day, maybe it will be.
 
-/* First used: 0 (reserved), Last used: 345 */
+   We used to tell people to "fix the above error[s] and try recompiling
+   the program" via a call to fatal, but that message tended to look
+   silly.  So instead, we just do the equivalent of a call to fatal in the
+   same situation (call exit).  */
+
+/* First used: 0 (reserved), Last used: 347 */
 
 void
 my_friendly_abort (i)
      int i;
 {
+  if (errorcount > 0 || sorrycount > 0)
+    exit (34);
+
   if (i == 0)
     error ("Internal compiler error.");
-  else if (errorcount + sorrycount == 1)
-    fatal ("please fix above error, and try recompiling.");
-  else if (errorcount > 0 || sorrycount > 0)
-    fatal ("please fix above errors, and try recompiling.");
   else
     error ("Internal compiler error %d.", i);
-  fatal ("Please report this to `bug-g++@prep.ai.mit.edu'.");
+
+  fatal ("Please submit a full bug report to `bug-g++@prep.ai.mit.edu'.");
 }
 
 void
@@ -310,12 +305,7 @@ my_friendly_assert (cond, where)
      int cond, where;
 {
   if (cond == 0)
-    {
-      /* Don't say "please fix above errors", since it's quite
-	 possible that we've lost somewhere unrelated to an error.  */
-      errorcount = sorrycount = 0;
-      my_friendly_abort (where);
-    }
+    my_friendly_abort (where);
 }
 
 /* Return nonzero if VALUE is a valid constant-valued expression
@@ -463,7 +453,7 @@ store_init_value (decl, init)
 
   /* Digest the specified initializer into an expression.  */
 
-  value = digest_init (type, init, 0);
+  value = digest_init (type, init, (tree *) 0);
 
   /* Store the expression if valid; else report error.  */
 
@@ -592,6 +582,13 @@ digest_init (type, init, tail)
 	  error ("union with no named members cannot be initialized");
 	  return error_mark_node;
 	}
+
+      if (raw_constructor && !TYPE_NEEDS_CONSTRUCTING (type))
+	{
+	  result = process_init_constructor (type, init, NULL_PTR);
+	  return result;
+	}
+
       if (! raw_constructor)
 	{
 	  error ("type mismatch in initialization");
@@ -609,7 +606,7 @@ digest_init (type, init, tail)
 	{
 	  /* Take just the first element from within the constructor
 	     and it should match the type of the first element.  */
-	  element = digest_init (TREE_TYPE (field), element, 0);
+	  element = digest_init (TREE_TYPE (field), element, (tree *) 0);
 	  result = build (CONSTRUCTOR, type, 0, build_tree_list (field, element));
 	  TREE_CONSTANT (result) = TREE_CONSTANT (element);
 	  TREE_STATIC (result) = (initializer_constant_valid_p (element)
@@ -754,7 +751,7 @@ process_init_constructor (type, init, elts)
   tree result;
   int allconstant = 1;
   int allsimple = 1;
-  int erred = 0;
+  int erroneous = 0;
 
   /* Make TAIL be the list of elements to use for the initialization,
      no matter how the data was given to us.  */
@@ -811,7 +808,7 @@ process_init_constructor (type, init, elts)
 	    }
 
 	  if (next1 == error_mark_node)
-	    erred = 1;
+	    erroneous = 1;
 	  else if (!TREE_CONSTANT (next1))
 	    allconstant = 0;
 	  else if (! initializer_constant_valid_p (next1))
@@ -857,7 +854,11 @@ process_init_constructor (type, init, elts)
 
 	  if (TREE_CODE (field) == CONST_DECL || TREE_CODE (field) == TYPE_DECL)
 	    continue;
-	  if (TREE_CODE (field) == VAR_DECL && !TREE_STATIC (field))
+
+	  /* A static mmmember isn't considered "part of the object", so
+	     it has no business even thinking about involving itself in
+	     what an initializer-list is trying to do.  */
+	  if (TREE_CODE (field) == VAR_DECL && TREE_STATIC (field))
 	    continue;
 
 	  if (TREE_VALUE (tail) != 0)
@@ -879,7 +880,7 @@ process_init_constructor (type, init, elts)
 	    }
 
 	  if (next1 == error_mark_node)
-	    erred = 1;
+	    erroneous = 1;
 	  else if (!TREE_CONSTANT (next1))
 	    allconstant = 0;
 	  else if (! initializer_constant_valid_p (next1))
@@ -896,7 +897,7 @@ process_init_constructor (type, init, elts)
 	    {
 	      register tree next1 = DECL_INITIAL (field);
 	      if (TREE_CODE (next1) == ERROR_MARK)
-		erred = 1;
+		erroneous = 1;
 	      else if (!TREE_CONSTANT (next1))
 		allconstant = 0;
 	      else if (! initializer_constant_valid_p (next1))
@@ -916,6 +917,77 @@ process_init_constructor (type, init, elts)
 	}
     }
 
+  if (TREE_CODE (type) == UNION_TYPE)
+    {
+      register tree field = TYPE_FIELDS (type);
+      register tree next1;
+
+      /* Find the first named field.  ANSI decided in September 1990
+	 that only named fields count here.  */
+      while (field && DECL_NAME (field) == 0)
+	field = TREE_CHAIN (field);
+
+      /* For a union, get the initializer for 1 fld.  */
+
+      if (tail == NULL_TREE)
+	{
+	  error ("empty initializer for union");
+	  tail = build_tree_list (NULL_TREE, NULL_TREE);
+	}
+
+      /* If this element specifies a field, initialize via that field.  */
+      if (TREE_PURPOSE (tail) != NULL_TREE)
+	{
+	  int win = 0;
+
+	  if (TREE_CODE (TREE_PURPOSE (tail)) == FIELD_DECL)
+	    /* Handle the case of a call by build_c_cast.  */
+	    field = TREE_PURPOSE (tail), win = 1;
+	  else if (TREE_CODE (TREE_PURPOSE (tail)) != IDENTIFIER_NODE)
+	    error ("index value instead of field name in union initializer");
+	  else
+	    {
+	      tree temp;
+	      for (temp = TYPE_FIELDS (type);
+		   temp;
+		   temp = TREE_CHAIN (temp))
+		if (DECL_NAME (temp) == TREE_PURPOSE (tail))
+		  break;
+	      if (temp)
+		field = temp, win = 1;
+	      else
+		error ("no field `%s' in union being initialized",
+		       IDENTIFIER_POINTER (TREE_PURPOSE (tail)));
+	    }
+	  if (!win)
+	    TREE_VALUE (tail) = error_mark_node;
+	}
+
+      if (TREE_VALUE (tail) != 0)
+	{
+	  tree tail1 = tail;
+
+	  next1 = digest_init (TREE_TYPE (field),
+			       TREE_VALUE (tail), &tail1);
+	  if (tail1 != 0 && TREE_CODE (tail1) != TREE_LIST)
+	    abort ();
+	  tail = tail1;
+	}
+      else
+	{
+	  next1 = error_mark_node;
+	  tail = TREE_CHAIN (tail);
+	}
+
+      if (next1 == error_mark_node)
+	erroneous = 1;
+      else if (!TREE_CONSTANT (next1))
+	allconstant = 0;
+      else if (initializer_constant_valid_p (next1) == 0)
+	allsimple = 0;
+      members = tree_cons (field, next1, members);
+    }
+
   /* If arguments were specified as a list, just remove the ones we used.  */
   if (elts)
     *elts = tail;
@@ -924,7 +996,7 @@ process_init_constructor (type, init, elts)
   else if (tail)
     warning ("excess elements in aggregate initializer");
 
-  if (erred)
+  if (erroneous)
     return error_mark_node;
 
   result = build (CONSTRUCTOR, type, NULL_TREE, nreverse (members));
@@ -969,7 +1041,7 @@ build_scoped_ref (datum, types)
      tree datum;
      tree types;
 {
-  tree orig_ref, ref;
+  tree ref;
   tree type = TREE_TYPE (datum);
 
   if (datum == error_mark_node)
@@ -980,7 +1052,7 @@ build_scoped_ref (datum, types)
     {
       /* We have some work to do.  */
       struct type_chain { tree type; struct type_chain *next; } *chain = 0, *head = 0, scratch;
-      orig_ref = ref = build_unary_op (ADDR_EXPR, datum, 0);
+      ref = build_unary_op (ADDR_EXPR, datum, 0);
       while (TREE_CODE (types) == SCOPE_REF)
 	{
 	  tree t = TREE_OPERAND (types, 1);
@@ -1011,7 +1083,7 @@ build_scoped_ref (datum, types)
 	      if (binfo == error_mark_node)
 		return error_mark_node;
 	      if (binfo == 0)
-		return error_not_base_type (TYPE_NAME_STRING (chain->type), TYPE_NAME_STRING (type));
+		return error_not_base_type (chain->type, type);
 	      ref = convert_pointer_to (binfo, ref);
 	    }
 	  chain = chain->next;
@@ -1079,7 +1151,7 @@ build_x_arrow (datum)
 
   if (IS_AGGR_TYPE (type) && TYPE_OVERLOADS_ARROW (type))
     {
-      while (rval = build_opfncall (COMPONENT_REF, LOOKUP_NORMAL, rval))
+      while (rval = build_opfncall (COMPONENT_REF, LOOKUP_NORMAL, rval, NULL_TREE, NULL_TREE))
 	{
 	  if (rval == error_mark_node)
 	    return error_mark_node;
@@ -1233,7 +1305,8 @@ build_functional_cast (exp, parms)
       && (TREE_CODE (IDENTIFIER_GLOBAL_VALUE (name)) == TREE_LIST
 	  || TREE_CODE (IDENTIFIER_GLOBAL_VALUE (name)) == FUNCTION_DECL))
     {
-      expr_as_fncall = build_overload_call (name, parms, 0, 0);
+      expr_as_fncall = build_overload_call (name, parms, 0,
+					    (struct candidate *)0);
       if (expr_as_fncall == NULL_TREE)
 	expr_as_fncall = error_mark_node;
     }
@@ -1425,7 +1498,8 @@ build_functional_cast (exp, parms)
 				NULL_TREE, LOOKUP_NORMAL);
     if (expr_as_fncall)
       return expr_as_fncall == error_mark_node
-	? build_overload_call (name, parms, 1, 0) : expr_as_fncall;
+	? build_overload_call (name, parms, 1, (struct candidate *)0)
+	  : expr_as_fncall;
     error ("invalid functional cast");
     return error_mark_node;
   }
