@@ -1,5 +1,5 @@
 /* Definitions for code generation pass of GNU compiler.
-   Copyright (C) 1987, 1991 Free Software Foundation, Inc.
+   Copyright (C) 1987, 1991, 1992, 1993 Free Software Foundation, Inc.
 
 This file is part of GNU CC.
 
@@ -179,7 +179,7 @@ enum direction {none, upward, downward};  /* Value has this type.  */
 #define FUNCTION_ARG_PADDING(MODE, TYPE)				\
   (((MODE) == BLKmode							\
     ? ((TYPE) && TREE_CODE (TYPE_SIZE (TYPE)) == INTEGER_CST		\
-       && int_size_in_bytes (TYPE) < PARM_BOUNDARY / BITS_PER_UNIT)	\
+       && int_size_in_bytes (TYPE) < (PARM_BOUNDARY / BITS_PER_UNIT))	\
     : GET_MODE_BITSIZE (MODE) < PARM_BOUNDARY)				\
    ? downward : upward)
 #else
@@ -202,19 +202,11 @@ enum direction {none, upward, downward};  /* Value has this type.  */
    - if the type is marked as addressable (it is required to be constructed
      into the stack)
    - if the padding and mode of the type is such that a copy into a register
-     would put it into the wrong part of the register
-   - when STRICT_ALIGNMENT and the type is BLKmode and is is not
-     aligned to a boundary corresponding to what can be loaded into a
-     register.  */
+     would put it into the wrong part of the register.
 
-#define MUST_PASS_IN_STACK_BAD_ALIGN(MODE,TYPE)			\
-  (STRICT_ALIGNMENT && MODE == BLKmode			        \
-   && TYPE_ALIGN (TYPE) < (BIGGEST_ALIGNMENT < BITS_PER_WORD	\
-			   ? BIGGEST_ALIGNMENT : BITS_PER_WORD))
-  
-/* Which padding can't be supported depends on the byte endianness.  */
+   Which padding can't be supported depends on the byte endianness.
 
-/* A value in a register is implicitly padded at the most significant end.
+   A value in a register is implicitly padded at the most significant end.
    On a big-endian machine, that is the lower end in memory.
    So a value padded in memory at the upper end can't go in a register.
    For a little-endian machine, the reverse is true.  */
@@ -230,16 +222,17 @@ enum direction {none, upward, downward};  /* Value has this type.  */
    && (TREE_CODE (TYPE_SIZE (TYPE)) != INTEGER_CST	\
        || TREE_ADDRESSABLE (TYPE)			\
        || ((MODE) == BLKmode 				\
+	   && ! ((TYPE) != 0 && TREE_CODE (TYPE_SIZE (TYPE)) == INTEGER_CST \
+		 && 0 == (int_size_in_bytes (TYPE)	\
+			  % (PARM_BOUNDARY / BITS_PER_UNIT))) \
 	   && (FUNCTION_ARG_PADDING (MODE, TYPE)	\
-	       == MUST_PASS_IN_STACK_BAD_PADDING))	\
-       || MUST_PASS_IN_STACK_BAD_ALIGN (MODE, TYPE)))
+	       == MUST_PASS_IN_STACK_BAD_PADDING))))
 
-/* Nonzero if type TYPE should be returned in memory
-   (even though its mode is not BLKmode).
+/* Nonzero if type TYPE should be returned in memory.
    Most machines can use the following default definition.  */
 
 #ifndef RETURN_IN_MEMORY
-#define RETURN_IN_MEMORY(TYPE) 0
+#define RETURN_IN_MEMORY(TYPE) (TYPE_MODE (TYPE) == BLKmode)
 #endif
 
 /* Optabs are tables saying how to generate insn bodies
@@ -329,6 +322,14 @@ extern optab sin_optab;		/* Sine */
 extern optab cos_optab;		/* Cosine */
 extern optab strlen_optab;	/* String length */
 
+/* Tables of patterns for extending one integer mode to another.  */
+extern enum insn_code extendtab[MAX_MACHINE_MODE][MAX_MACHINE_MODE][2];
+
+/* Tables of patterns for converting between fixed and floating point. */
+extern enum insn_code fixtab[NUM_MACHINE_MODES][NUM_MACHINE_MODES][2];
+extern enum insn_code fixtrunctab[NUM_MACHINE_MODES][NUM_MACHINE_MODES][2];
+extern enum insn_code floattab[NUM_MACHINE_MODES][NUM_MACHINE_MODES][2];
+
 /* Passed to expand_binop and expand_unop to say which options to try to use
    if the requested operation can't be open-coded on the requisite mode.
    Either OPTAB_LIB or OPTAB_LIB_WIDEN says try using a library call.
@@ -455,228 +456,355 @@ extern rtxfun bcc_gen_fctn[NUM_RTX_CODE];
 
 extern enum insn_code setcc_gen_code[NUM_RTX_CODE];
 
+/* This array records the insn_code of insns to perform block moves.  */
+extern enum insn_code movstr_optab[NUM_MACHINE_MODES];
+
+/* Define functions given in optabs.c.  */
+
 /* Expand a binary operation given optab and rtx operands.  */
-extern rtx expand_binop ();
+extern rtx expand_binop PROTO((enum machine_mode, optab, rtx, rtx, rtx,
+			       int, enum optab_methods));
 
 /* Expand a binary operation with both signed and unsigned forms.  */
-extern rtx sign_expand_binop ();
+extern rtx sign_expand_binop PROTO((enum machine_mode, optab, optab, rtx,
+				    rtx, rtx, int, enum optab_methods));
+
+/* Generate code to perform an operation on two operands with two results.  */
+extern int expand_twoval_binop PROTO((optab, rtx, rtx, rtx, rtx, int));
 
 /* Expand a unary arithmetic operation given optab rtx operand.  */
-extern rtx expand_unop ();
+extern rtx expand_unop PROTO((enum machine_mode, optab, rtx, rtx, int));
 
 /* Expand the complex absolute value operation.  */
-extern rtx expand_complex_abs ();
+extern rtx expand_complex_abs PROTO((enum machine_mode, rtx, rtx, int));
 
-/* Arguments MODE, RTX: return an rtx for the negation of that value.
-   May emit insns.  */
-extern rtx negate_rtx ();
+/* Generate an instruction with a given INSN_CODE with an output and
+   an input.  */
+extern void emit_unop_insn PROTO((int, rtx, rtx, enum rtx_code));
 
-/* Expand a logical AND operation.  */
-extern rtx expand_and ();
+/* Emit code to perform a series of operations on a multi-word quantity, one
+   word at a time.  */
+extern rtx emit_no_conflict_block PROTO((rtx, rtx, rtx, rtx, rtx));
 
-/* Emit a store-flag operation.  */
-extern rtx emit_store_flag ();
+/* Emit code to make a call to a constant function or a library call. */
+extern void emit_libcall_block PROTO((rtx, rtx, rtx, rtx));
 
-/* Return the CODE_LABEL rtx for a LABEL_DECL, creating it if necessary.  */
-extern rtx label_rtx ();
+/* Emit one rtl instruction to store zero in specified rtx.  */
+extern void emit_clr_insn PROTO((rtx));
 
-/* Given a JUMP_INSN, return a description of the test being made.  */
-extern rtx get_condition ();
+/* Emit one rtl insn to store 1 in specified rtx assuming it contains 0.  */
+extern void emit_0_to_1_insn PROTO((rtx));
 
-/* Return the INSN_CODE to use for an extend operation.  */
-extern enum insn_code can_extend_p ();
+/* Emit one rtl insn to compare two rtx's.  */
+extern void emit_cmp_insn PROTO((rtx, rtx, enum rtx_code, rtx,
+				 enum machine_mode, int, int));
 
-/* Initialize the tables that control conversion between fixed and
-   floating values.  */
-extern void init_fixtab ();
-extern void init_floattab ();
+/* Nonzero if a compare of mode MODE can be done straightforwardly
+   (without splitting it into pieces).  */
+extern int can_compare_p PROTO((enum machine_mode));
 
-/* Generate code for a FIX_EXPR.  */
-extern void expand_fix ();
-
-/* Generate code for a FLOAT_EXPR.  */
-extern void expand_float ();
+/* Generate code to indirectly jump to a location given in the rtx LOC.  */
+extern void emit_indirect_jump PROTO((rtx));
 
 /* Create but don't emit one rtl instruction to add one rtx into another.
    Modes must match; operands must meet the operation's predicates.
    Likewise for subtraction and for just copying.
    These do not call protect_from_queue; caller must do so.  */
-extern rtx gen_add2_insn ();
-extern rtx gen_sub2_insn ();
-extern rtx gen_move_insn ();
+extern rtx gen_add2_insn PROTO((rtx, rtx));
+extern rtx gen_sub2_insn PROTO((rtx, rtx));
+extern rtx gen_move_insn PROTO((rtx, rtx));
+extern int have_add2_insn PROTO((enum machine_mode));
+extern int have_sub2_insn PROTO((enum machine_mode));
 
-/* Emit one rtl instruction to store zero in specified rtx.  */
-extern void emit_clr_insn ();
+/* Return the INSN_CODE to use for an extend operation.  */
+extern enum insn_code can_extend_p PROTO((enum machine_mode,
+					  enum machine_mode, int));
 
-/* Emit one rtl insn to store 1 in specified rtx assuming it contains 0.  */
-extern void emit_0_to_1_insn ();
+/* Generate the body of an insn to extend Y (with mode MFROM)
+   into X (with mode MTO).  Do zero-extension if UNSIGNEDP is nonzero.  */
+extern rtx gen_extend_insn PROTO((rtx, rtx, enum machine_mode,
+				  enum machine_mode, int));
 
-/* Emit one rtl insn to compare two rtx's.  */
-extern void emit_cmp_insn ();
+/* Initialize the tables that control conversion between fixed and
+   floating values.  */
+extern void init_fixtab PROTO((void));
+extern void init_floattab PROTO((void));
 
-/* Generate rtl to compare two rtx's, will call emit_cmp_insn.  */
-extern rtx compare_from_rtx ();
+/* Generate code for a FLOAT_EXPR.  */
+extern void expand_float PROTO((rtx, rtx, int));
+
+/* Generate code for a FIX_EXPR.  */
+extern void expand_fix PROTO((rtx, rtx, int));
+
+/* Call this once to initialize the contents of the optabs
+   appropriately for the current target machine.  */
+extern void init_optabs	PROTO((void));
+
+/* Functions from expmed.c:  */
+
+/* Arguments MODE, RTX: return an rtx for the negation of that value.
+   May emit insns.  */
+extern rtx negate_rtx PROTO((enum machine_mode, rtx));
+
+/* Expand a logical AND operation.  */
+extern rtx expand_and PROTO((rtx, rtx, rtx));
+
+/* Emit a store-flag operation.  */
+extern rtx emit_store_flag PROTO((rtx, enum rtx_code, rtx, rtx,
+				  enum machine_mode, int, int));
+
+/* Functions from loop.c:  */
+
+/* Given a JUMP_INSN, return a description of the test being made.  */
+extern rtx get_condition PROTO((rtx, rtx *));
+
+/* Functions from expr.c:  */
+
+/* This is run once per compilation to set up which modes can be used
+   directly in memory and to initialize the block move optab.  */
+extern void init_expr_once PROTO((void));
+
+/* This is run at the start of compiling a function.  */
+extern void init_expr PROTO((void));
+
+/* Use protect_from_queue to convert a QUEUED expression
+   into something that you can put immediately into an instruction.  */
+extern rtx protect_from_queue PROTO((rtx, int));
+
+/* Perform all the pending incrementations.  */
+extern void emit_queue PROTO((void));
 
 /* Emit some rtl insns to move data between rtx's, converting machine modes.
    Both modes must be floating or both fixed.  */
-extern void convert_move ();
+extern void convert_move PROTO((rtx, rtx, int));
 
 /* Convert an rtx to specified machine mode and return the result.  */
-extern rtx convert_to_mode ();
+extern rtx convert_to_mode PROTO((enum machine_mode, rtx, int));
 
-/* Emit code to push some arguments and call a library routine,
-   storing the value in a specified place.  Calling sequence is
-   complicated.  */
-extern void emit_library_call ();
+/* Emit code to move a block Y to a block X.  */
+extern void emit_block_move PROTO((rtx, rtx, rtx, int));
 
-/* Given an rtx that may include add and multiply operations,
-   generate them as insns and return a pseudo-reg containing the value.
-   Useful after calling expand_expr with 1 as sum_ok.  */
-extern rtx force_operand ();
+/* Copy all or part of a value X into registers starting at REGNO.
+   The number of registers to be filled is NREGS.  */
+extern void move_block_to_reg PROTO((int, rtx, int, enum machine_mode));
 
-/* Return an rtx for the size in bytes of the value of an expr.  */
-extern rtx expr_size ();
+/* Copy all or part of a BLKmode value X out of registers starting at REGNO.
+   The number of registers to be filled is NREGS.  */
+extern void move_block_from_reg PROTO((int, rtx, int));
 
-extern rtx lookup_static_chain ();
+/* Mark NREGS consecutive regs, starting at REGNO, as being live now.  */
+extern void use_regs PROTO((int, int));
 
-/* Return an rtx like arg but sans any constant terms.
-   Returns the original rtx if it has no constant terms.
-   The constant terms are added and stored via a second arg.  */
-extern rtx eliminate_constant_term ();
+/* Write zeros through the storage of OBJECT.
+   If OBJECT has BLKmode, SIZE is its length in bytes.  */
+extern void clear_storage PROTO((rtx, int));
 
-/* Convert arg to a valid memory address for specified machine mode,
-   by emitting insns to perform arithmetic if nec.  */
-extern rtx memory_address ();
+/* Emit insns to set X from Y.  */
+extern rtx emit_move_insn PROTO((rtx, rtx));
 
-/* Like `memory_address' but pretent `flag_force_addr' is 0.  */
-extern rtx memory_address_noforce ();
-
-/* Return a memory reference like MEMREF, but with its mode changed
-   to MODE and its address changed to ADDR.
-   (VOIDmode means don't change the mode.
-   NULL for ADDR means don't change the address.)  */
-extern rtx change_address ();
-
-/* Return a memory reference like MEMREF, but which is known to have a
-   valid address.  */
-
-extern rtx validize_mem ();
-
-/* Convert a stack slot address ADDR valid in function FNDECL
-   into an address valid in this function (using a static chain).  */
-extern rtx fix_lexical_addr ();
-
-/* Return the address of the trampoline for entering nested fn FUNCTION.  */
-extern rtx trampoline_address ();
-
-/* Assemble the static constant template for function entry trampolines.  */
-extern rtx assemble_trampoline_template ();
-
-/* Return 1 if two rtx's are equivalent in structure and elements.  */
-extern int rtx_equal_p ();
-
-/* Given rtx, return new rtx whose address won't be affected by
-   any side effects.  It has been copied to a new temporary reg.  */
-extern rtx stabilize ();
-
-/* Given an rtx, copy all regs it refers to into new temps
-   and return a modified copy that refers to the new temps.  */
-extern rtx copy_all_regs ();
-
-/* Copy given rtx to a new temp reg and return that.  */
-extern rtx copy_to_reg ();
-
-/* Like copy_to_reg but always make the reg Pmode.  */
-extern rtx copy_addr_to_reg ();
-
-/* Like copy_to_reg but always make the reg the specified mode MODE.  */
-extern rtx copy_to_mode_reg ();
-
-/* Copy given rtx to given temp reg and return that.  */
-extern rtx copy_to_suggested_reg ();
-
-/* Copy a value to a register if it isn't already a register.
-   Args are mode (in case value is a constant) and the value.  */
-extern rtx force_reg ();
-
-/* Return given rtx, copied into a new temp reg if it was in memory.  */
-extern rtx force_not_mem ();
-
-/* Remove some bytes from the stack.  An rtx says how many.  */
-extern void adjust_stack ();
-
-/* Add some bytes to the stack.  An rtx says how many.  */
-extern void anti_adjust_stack ();
-
-/* This enum is used for the following two functions.  */
-enum save_level {SAVE_BLOCK, SAVE_FUNCTION, SAVE_NONLOCAL};
-
-/* Save the stack pointer at the specified level.  */
-extern void emit_stack_save ();
-
-/* Restore the stack pointer from a save area of the specified level.  */
-extern void emit_stack_restore ();
-
-/* Allocate some space on the stack dynamically and return its address.  An rtx
-   says how many bytes.  */
-extern rtx allocate_dynamic_stack_space ();
-
-/* Emit code to copy function value to a new temp reg and return that reg.  */
-extern rtx function_value ();
-
-/* Return an rtx that refers to the value returned by a function
-   in its original home.  This becomes invalid if any more code is emitted.  */
-extern rtx hard_function_value ();
-
-/* Return an rtx that refers to the value returned by a library call
-   in its original home.  This becomes invalid if any more code is emitted.  */
-extern rtx hard_libcall_value ();
-
-/* Emit code to copy function value to a specified place.  */
-extern void copy_function_value ();
-
-/* Given an rtx, return an rtx for a value rounded up to a multiple
-   of STACK_BOUNDARY / BITS_PER_UNIT.  */
-extern rtx round_push ();
+/* Emit insns to set X from Y, with no frills.  */
+extern rtx emit_move_insn_1 PROTO ((rtx, rtx));
 
 /* Push a block of length SIZE (perhaps variable)
    and return an rtx to address the beginning of the block.  */
-extern rtx push_block ();
+extern rtx push_block PROTO((rtx, int, int));
+
+/* Make an operand to push someting on the stack.  */
+extern rtx gen_push_operand PROTO((void));
+
+#ifdef TREE_CODE
+/* Generate code to push something onto the stack, given its mode and type.  */
+extern void emit_push_insn PROTO((rtx, enum machine_mode, tree, rtx, int,
+				  int, rtx, int, rtx, rtx));
+
+/* Emit library call.  These cannot have accurate prototypes since they have
+   a variable number of args.  */
+extern void emit_library_call ();
+extern void emit_library_call_value ();
+
+/* Expand an assignment that stores the value of FROM into TO. */
+extern rtx expand_assignment PROTO((tree, tree, int, int));
 
 /* Generate code for computing expression EXP,
    and storing the value into TARGET.
    If SUGGEST_REG is nonzero, copy the value through a register
    and return that register, if that is possible.  */
-extern rtx store_expr ();
+extern rtx store_expr PROTO((tree, rtx, int));
+#endif
 
-extern rtx prepare_call_address ();
-extern rtx expand_call ();
-extern void emit_call_1 ();
+/* Given an rtx that may include add and multiply operations,
+   generate them as insns and return a pseudo-reg containing the value.
+   Useful after calling expand_expr with 1 as sum_ok.  */
+extern rtx force_operand PROTO((rtx, rtx));
 
-extern void emit_block_move ();
-extern void emit_push_insn ();
-extern void use_regs ();
-extern void move_block_to_reg ();
+#ifdef TREE_CODE
+/* Generate code for computing expression EXP.
+   An rtx for the computed value is returned.  The value is never null.
+   In the case of a void EXP, const0_rtx is returned.  */
+extern rtx expand_expr PROTO((tree, rtx, enum machine_mode,
+			      enum expand_modifier));
+#endif
 
-extern rtx store_bit_field ();
-extern rtx extract_bit_field ();
-extern rtx expand_shift ();
-extern rtx expand_mult ();
-extern rtx expand_divmod ();
-extern rtx expand_mult_add ();
-extern rtx expand_stmt_expr ();
-extern rtx emit_no_conflict_block ();
-extern void emit_libcall_block ();
+/* At the start of a function, record that we have no previously-pushed
+   arguments waiting to be popped.  */
+extern void init_pending_stack_adjust PROTO((void));
 
-extern void jumpifnot ();
-extern void jumpif ();
-extern void do_jump ();
+/* When exiting from function, if safe, clear out any pending stack adjust
+   so the adjustment won't get done.  */
+extern void clear_pending_stack_adjust PROTO((void));
 
-extern rtx assemble_static_space ();
+/* Pop any previously-pushed arguments that have not been popped yet.  */
+extern void do_pending_stack_adjust PROTO((void));
 
-extern void locate_and_pad_parm ();
+#ifdef TREE_CODE
+/* Expand all cleanups up to OLD_CLEANUPS.  */
+extern void expand_cleanups_to PROTO((tree));
 
-extern rtx expand_inline_function ();
+/* Generate code to evaluate EXP and jump to LABEL if the value is zero.  */
+extern void jumpifnot PROTO((tree, rtx));
+
+/* Generate code to evaluate EXP and jump to LABEL if the value is nonzero.  */
+extern void jumpif PROTO((tree, rtx));
+
+/* Generate code to evaluate EXP and jump to IF_FALSE_LABEL if
+   the result is zero, or IF_TRUE_LABEL if the result is one.  */
+extern void do_jump PROTO((tree, rtx, rtx));
+#endif
+
+/* Generate rtl to compare two rtx's, will call emit_cmp_insn.  */
+extern rtx compare_from_rtx PROTO((rtx, rtx, enum rtx_code, int,
+				   enum machine_mode, rtx, int));
+
+/* Generate a tablejump instruction (used for switch statements).  */
+extern void do_tablejump PROTO((rtx, enum machine_mode, rtx, rtx, rtx));
+
+#ifdef TREE_CODE
+/* rtl.h and tree.h were included.  */
+/* Return an rtx for the size in bytes of the value of an expr.  */
+extern rtx expr_size PROTO((tree));
+
+extern rtx lookup_static_chain PROTO((tree));
+
+/* Convert a stack slot address ADDR valid in function FNDECL
+   into an address valid in this function (using a static chain).  */
+extern rtx fix_lexical_addr PROTO((rtx, tree));
+
+/* Return the address of the trampoline for entering nested fn FUNCTION.  */
+extern rtx trampoline_address PROTO((tree));
+
+/* Return an rtx that refers to the value returned by a function
+   in its original home.  This becomes invalid if any more code is emitted.  */
+extern rtx hard_function_value PROTO((tree, tree));
+
+extern rtx prepare_call_address	PROTO((rtx, tree, rtx *));
+
+extern rtx expand_call PROTO((tree, rtx, int));
+
+extern rtx expand_shift PROTO((enum tree_code, enum machine_mode, rtx, tree, rtx, int));
+extern rtx expand_divmod PROTO((int, enum tree_code, enum machine_mode, rtx, rtx, rtx, int));
+extern void locate_and_pad_parm PROTO((enum machine_mode, tree, int, tree, struct args_size *, struct args_size *, struct args_size *));
+extern rtx expand_inline_function PROTO((tree, tree, rtx, int, tree, rtx));
+/* Return the CODE_LABEL rtx for a LABEL_DECL, creating it if necessary.  */
+extern rtx label_rtx PROTO((tree));
+#endif
+
+/* Indicate how an input argument register was promoted.  */
+extern rtx promoted_input_arg PROTO((int, enum machine_mode *, int *));
+
+/* Return an rtx like arg but sans any constant terms.
+   Returns the original rtx if it has no constant terms.
+   The constant terms are added and stored via a second arg.  */
+extern rtx eliminate_constant_term PROTO((rtx, rtx *));
+
+/* Convert arg to a valid memory address for specified machine mode,
+   by emitting insns to perform arithmetic if nec.  */
+extern rtx memory_address PROTO((enum machine_mode, rtx));
+
+/* Like `memory_address' but pretent `flag_force_addr' is 0.  */
+extern rtx memory_address_noforce PROTO((enum machine_mode, rtx));
+
+/* Return a memory reference like MEMREF, but with its mode changed
+   to MODE and its address changed to ADDR.
+   (VOIDmode means don't change the mode.
+   NULL for ADDR means don't change the address.)  */
+extern rtx change_address PROTO((rtx, enum machine_mode, rtx));
+
+/* Return a memory reference like MEMREF, but which is known to have a
+   valid address.  */
+
+extern rtx validize_mem PROTO((rtx));
+
+/* Assemble the static constant template for function entry trampolines.  */
+extern rtx assemble_trampoline_template PROTO((void));
+
+/* Return 1 if two rtx's are equivalent in structure and elements.  */
+extern int rtx_equal_p PROTO((rtx, rtx));
+
+/* Given rtx, return new rtx whose address won't be affected by
+   any side effects.  It has been copied to a new temporary reg.  */
+extern rtx stabilize PROTO((rtx));
+
+/* Given an rtx, copy all regs it refers to into new temps
+   and return a modified copy that refers to the new temps.  */
+extern rtx copy_all_regs PROTO((rtx));
+
+/* Copy given rtx to a new temp reg and return that.  */
+extern rtx copy_to_reg PROTO((rtx));
+
+/* Like copy_to_reg but always make the reg Pmode.  */
+extern rtx copy_addr_to_reg PROTO((rtx));
+
+/* Like copy_to_reg but always make the reg the specified mode MODE.  */
+extern rtx copy_to_mode_reg PROTO((enum machine_mode, rtx));
+
+/* Copy given rtx to given temp reg and return that.  */
+extern rtx copy_to_suggested_reg PROTO((rtx, rtx, enum machine_mode));
+
+/* Copy a value to a register if it isn't already a register.
+   Args are mode (in case value is a constant) and the value.  */
+extern rtx force_reg PROTO((enum machine_mode, rtx));
+
+/* Return given rtx, copied into a new temp reg if it was in memory.  */
+extern rtx force_not_mem PROTO((rtx));
+
+/* Remove some bytes from the stack.  An rtx says how many.  */
+extern void adjust_stack PROTO((rtx));
+
+/* Add some bytes to the stack.  An rtx says how many.  */
+extern void anti_adjust_stack PROTO((rtx));
+
+/* This enum is used for the following two functions.  */
+enum save_level {SAVE_BLOCK, SAVE_FUNCTION, SAVE_NONLOCAL};
+
+/* Save the stack pointer at the specified level.  */
+extern void emit_stack_save PROTO((enum save_level, rtx *, rtx));
+
+/* Restore the stack pointer from a save area of the specified level.  */
+extern void emit_stack_restore PROTO((enum save_level, rtx, rtx));
+
+/* Allocate some space on the stack dynamically and return its address.  An rtx
+   says how many bytes.  */
+extern rtx allocate_dynamic_stack_space PROTO((rtx, rtx, int));
+
+/* Emit code to copy function value to a new temp reg and return that reg.  */
+extern rtx function_value ();
+
+/* Return an rtx that refers to the value returned by a library call
+   in its original home.  This becomes invalid if any more code is emitted.  */
+extern rtx hard_libcall_value PROTO((enum machine_mode));
+
+/* Given an rtx, return an rtx for a value rounded up to a multiple
+   of STACK_BOUNDARY / BITS_PER_UNIT.  */
+extern rtx round_push PROTO((rtx));
+
+extern void emit_block_move PROTO((rtx, rtx, rtx, int));
+
+extern rtx store_bit_field PROTO((rtx, int, int, enum machine_mode, rtx, int, int));
+extern rtx extract_bit_field PROTO((rtx, int, int, int, rtx, enum machine_mode, enum machine_mode, int, int));
+extern rtx expand_mult PROTO((enum machine_mode, rtx, rtx, rtx, int));
+extern rtx expand_mult_add PROTO((rtx, rtx, rtx, rtx,enum machine_mode, int));
+
+extern rtx assemble_static_space PROTO((int));
 
 /* Hook called by expand_expr for language-specific tree codes.
    It is up to the language front end to install a hook
