@@ -73,6 +73,10 @@ extern int extra_warnings;
 
 extern int warn_unused;
 
+/* Nonzero means warn if inline function is too large.  */
+
+extern int warn_inline;
+
 /* Nonzero to warn about variables used before they are initialized.  */
 
 extern int warn_uninitialized;
@@ -96,6 +100,11 @@ extern int warn_return_type;
 
 extern int warn_cast_align;
 
+/* Nonzero means warn that dbx info for template class methods isn't fully
+   supported yet.  */
+
+extern int warn_template_debugging;
+
 /* Nonzero means warn about any identifiers that match in the first N
    characters.  The value N is in `id_clash_len'.  */
 
@@ -224,6 +233,10 @@ extern int flag_no_peephole;
 
 extern int flag_volatile;
 
+/* Nonzero means treat all global and extern variables as global.  */
+
+extern int flag_volatile_global;
+
 /* Nonzero allows GCC to violate some IEEE or ANSI rules regarding math
    operations in the interest of optimization.  For example it allows
    GCC to assume arguments to sqrt are nonnegative numbers, allowing
@@ -237,7 +250,7 @@ extern int flag_fast_math;
 extern int flag_inline_functions;
 
 /* Nonzero for -fkeep-inline-functions: even if we make a function
-   go inline everywhere, keep its defintion around for debugging
+   go inline everywhere, keep its definition around for debugging
    purposes.  */
 
 extern int flag_keep_inline_functions;
