@@ -429,7 +429,7 @@ from the machine description file `md'.  */\n\n");
   printf ("extern rtx *recog_operand_loc[];\n");
   printf ("extern rtx *recog_dup_loc[];\n");
   printf ("extern char recog_dup_num[];\n");
-  printf ("extern\n#ifdef __GNUC__\nvolatile\n#endif\n");
+  printf ("extern\n#ifdef __GNUC__\n__volatile__\n#endif\n");
   printf ("void fatal_insn_not_found ();\n\n");
 
   printf ("void\ninsn_extract (insn)\n");
@@ -504,7 +504,7 @@ from the machine description file `md'.  */\n\n");
 	  if (p->oplocs[i] == 0)
 	    {
 	      printf ("      ro[%d] = const0_rtx;\n", i);
-	      printf ("      ro_loc[%d] = &junk;\n", i, i);
+	      printf ("      ro_loc[%d] = &junk;\n", i);
 	    }
 	  else
 	    {
