@@ -65,6 +65,17 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include "recog.h"
 #include "output.h"
 
+/* Pseudos allocated here cannot be reallocated by global.c if the hard
+   register is used as a spill register.  So we don't allocate such pseudos
+   here if their preferred class is likely to be used by spills.
+
+   On most machines, the appropriate test is if the class has one
+   register, so we default to that.  */
+
+#ifndef CLASS_LIKELY_SPILLED_P
+#define CLASS_LIKELY_SPILLED_P(CLASS) (reg_class_size[(int) (CLASS)] == 1)
+#endif
+
 /* Next quantity number available for allocation.  */
 
 static int next_qty;
@@ -105,7 +116,7 @@ static char *qty_phys_has_sugg;
 
 /* Element Q is the number of refs to quantity Q.  */
 
-static short *qty_n_refs;
+static int *qty_n_refs;
 
 /* Element Q is a reg class contained in (smaller than) the
    preferred classes of all the pseudo regs that are tied in quantity Q.
@@ -159,13 +170,13 @@ static rtx *qty_scratch_rtx;
    reg_qty value is Q, or -1 is this quantity is for a SCRATCH.  This
    register should be the head of the chain maintained in reg_next_in_qty.  */
 
-static short *qty_first_reg;
+static int *qty_first_reg;
 
 /* If (REG N) has been assigned a quantity number, is a register number
    of another register assigned the same quantity number, or -1 for the
    end of the chain.  qty_first_reg point to the head of this chain.  */
 
-static short *reg_next_in_qty;
+static int *reg_next_in_qty;
 
 /* reg_qty[N] (where N is a pseudo reg number) is the qty number of that reg
    if it is >= 0,
@@ -329,11 +340,11 @@ alloc_qty_for_scratch (scratch, n, insn, insn_code_num, insn_number)
 	break;
       }
 
-  /* If CLASS has only one register, don't allocate the SCRATCH here since
+  /* If CLASS has only a few registers, don't allocate the SCRATCH here since
      it will prevent that register from being used as a spill register.
      reload will do the allocation.  */
 
-  if (class == NO_REGS || reg_class_size[(int) class] == 1)
+  if (class == NO_REGS || CLASS_LIKELY_SPILLED_P (class))
     return;
 
 #else /* REGISTER_CONSTRAINTS */
@@ -392,17 +403,17 @@ local_alloc ()
   qty_birth = (int *) alloca (max_qty * sizeof (int));
   qty_death = (int *) alloca (max_qty * sizeof (int));
   qty_scratch_rtx = (rtx *) alloca (max_qty * sizeof (rtx));
-  qty_first_reg = (short *) alloca (max_qty * sizeof (short));
+  qty_first_reg = (int *) alloca (max_qty * sizeof (int));
   qty_size = (int *) alloca (max_qty * sizeof (int));
   qty_mode = (enum machine_mode *) alloca (max_qty * sizeof (enum machine_mode));
   qty_n_calls_crossed = (int *) alloca (max_qty * sizeof (int));
   qty_min_class = (enum reg_class *) alloca (max_qty * sizeof (enum reg_class));
   qty_alternate_class = (enum reg_class *) alloca (max_qty * sizeof (enum reg_class));
-  qty_n_refs = (short *) alloca (max_qty * sizeof (short));
+  qty_n_refs = (int *) alloca (max_qty * sizeof (int));
 
   reg_qty = (int *) alloca (max_regno * sizeof (int));
   reg_offset = (char *) alloca (max_regno * sizeof (char));
-  reg_next_in_qty = (short *) alloca (max_regno * sizeof (short));
+  reg_next_in_qty = (int *) alloca (max_regno * sizeof (int));
 
   reg_renumber = (short *) oballoc (max_regno * sizeof (short));
   for (i = 0; i < max_regno; i++)
@@ -411,7 +422,7 @@ local_alloc ()
   /* Determine which pseudo-registers can be allocated by local-alloc.
      In general, these are the registers used only in a single block and
      which only die once.  However, if a register's preferred class has only
-     one entry, don't allocate this register here unless it is preferred
+     a few entries, don't allocate this register here unless it is preferred
      or nothing since retry_global_alloc won't be able to move it to
      GENERAL_REGS if a reload register of this class is needed.
 
@@ -422,7 +433,7 @@ local_alloc ()
     {
       if (reg_basic_block[i] >= 0 && reg_n_deaths[i] == 1
 	  && (reg_alternate_class (i) == NO_REGS
-	      || reg_class_size[(int) reg_preferred_class (i)] > 1))
+	      || ! CLASS_LIKELY_SPILLED_P (reg_preferred_class (i))))
 	reg_qty[i] = -2;
       else
 	reg_qty[i] = -1;
@@ -697,12 +708,6 @@ optimize_reg_copy_1 (insn, dest, src)
 	  int n_calls = 0;
 	  int d_n_calls = 0;
 
-	  /* If P is a CALL_INSN, SRC crosses one more call, since it
-	     used to die there.  */
-
-	  if (GET_CODE (p) == CALL_INSN)
-	    n_calls++;
-
 	  /* We can do the optimization.  Scan forward from INSN again,
 	     replacing regs as we go.  Set FAILED if a replacement can't
 	     be done.  In that case, we can't move the death note for SRC.
@@ -1079,7 +1084,7 @@ block_alloc (b)
   int insn_number = 0;
   int insn_count = 0;
   int max_uid = get_max_uid ();
-  short *qty_order;
+  int *qty_order;
   int no_conflict_combined_regno = -1;
 
   /* Count the instructions in the basic block.  */
@@ -1142,7 +1147,15 @@ block_alloc (b)
 	     Suitable insns are those with at least two operands and where
 	     operand 0 is an output that is a register that is not
 	     earlyclobber.
-	     For a commutative operation, try (set reg0 (arithop ... reg1)).
+
+	     We can tie operand 0 with some operand that dies in this insn.
+	     First look for operands that are required to be in the same
+	     register as operand 0.  If we find such, only try tying that
+	     operand or one that can be put into that operand if the
+	     operation is commutative.  If we don't find an operand
+	     that is required to be in the same register as operand 0,
+	     we can tie with any operand.
+
 	     Subregs in place of regs are also ok.
 
 	     If tying is done, WIN is set nonzero.  */
@@ -1158,56 +1171,64 @@ block_alloc (b)
 #endif
 	      )
 	    {
+#ifdef REGISTER_CONSTRAINTS
+	      int must_match_0 = -1;
+
+
+	      for (i = 1; i < insn_n_operands[insn_code_number]; i++)
+		if (requires_inout_p
+		    (insn_operand_constraint[insn_code_number][i]))
+		  must_match_0 = i;
+#endif
+
 	      r0 = recog_operand[0];
-	      r1 = recog_operand[1];
-
-	      /* If the first operand is an address, find a register in it.
-		 There may be more than one register, but we only try one of
-		 them.  */
-	      if (
-#ifdef REGISTER_CONSTRAINTS
-		  insn_operand_constraint[insn_code_number][1][0] == 'p'
-#else
-		  insn_operand_address_p[insn_code_number][1]
-#endif
-		  )
-		while (GET_CODE (r1) == PLUS || GET_CODE (r1) == MULT)
-		  r1 = XEXP (r1, 0);
-
-	      if (GET_CODE (r0) == REG || GET_CODE (r0) == SUBREG)
+	      for (i = 1; i < insn_n_operands[insn_code_number]; i++)
 		{
-		  /* We have two priorities for hard register preferences.
-		     If we have a move insn or an insn whose first input can
-		     only be in the same register as the output, give
-		     priority to an equivalence found from that insn.  */
 #ifdef REGISTER_CONSTRAINTS
-		  int may_save_copy
-		    = ((SET_DEST (body) == r0 && SET_SRC (body) == r1)
-		       || (r1 == recog_operand[1]
-			   && (requires_inout_p (insn_operand_constraint[insn_code_number][1]))));
-#else
-		  int may_save_copy = 0;
+		  /* Skip this operand if we found an operand that
+		     must match operand 0 and this operand isn't it
+		     and can't be made to be it by commutativity.  */
+
+		  if (must_match_0 >= 0 && i != must_match_0
+		      && ! (i == must_match_0 + 1
+			    && insn_operand_constraint[insn_code_number][i-1][0] == '%')
+		      && ! (i == must_match_0 - 1
+			    && insn_operand_constraint[insn_code_number][i][0] == '%'))
+		    continue;
 #endif
 
-		  if (GET_CODE (r1) == REG || GET_CODE (r1) == SUBREG)
-		    win = combine_regs (r1, r0, may_save_copy,
-					insn_number, insn, 0);
+		  r1 = recog_operand[i];
 
-		  if (win == 0
-		      && insn_n_operands[insn_code_number] > 2
+		  /* If the operand is an address, find a register in it.
+		     There may be more than one register, but we only try one
+		     of them.  */
+		  if (
 #ifdef REGISTER_CONSTRAINTS
-		      && insn_operand_constraint[insn_code_number][1][0] == '%'
+		      insn_operand_constraint[insn_code_number][i][0] == 'p'
 #else
-		      && GET_CODE (PATTERN (insn)) == SET
-		      && (GET_RTX_CLASS (GET_CODE (SET_SRC (PATTERN (insn))))
-			  == 'c')
-		      && rtx_equal_p (recog_operand[2],
-				      XEXP (SET_SRC (PATTERN (insn)), 0))
+		      insn_operand_address_p[insn_code_number][i]
 #endif
-		      && (r1 = recog_operand[2],
-			  GET_CODE (r1) == REG || GET_CODE (r1) == SUBREG))
-		    win = combine_regs (r1, r0, may_save_copy,
-					insn_number, insn, 0);
+		      )
+		    while (GET_CODE (r1) == PLUS || GET_CODE (r1) == MULT)
+		      r1 = XEXP (r1, 0);
+
+		  if (GET_CODE (r0) == REG || GET_CODE (r0) == SUBREG)
+		    {
+		      /* We have two priorities for hard register preferences.
+			 If we have a move insn or an insn whose first input
+			 can only be in the same register as the output, give
+			 priority to an equivalence found from that insn.  */
+		      int may_save_copy
+			= ((SET_DEST (body) == r0 && SET_SRC (body) == r1)
+#ifdef REGISTER_CONSTRAINTS
+			   || (r1 == recog_operand[i] && must_match_0 >= 0)
+#endif
+			   );
+		      
+		      if (GET_CODE (r1) == REG || GET_CODE (r1) == SUBREG)
+			win = combine_regs (r1, r0, may_save_copy,
+					    insn_number, insn, 0);
+		    }
 		}
 	    }
 
@@ -1229,6 +1250,7 @@ block_alloc (b)
 	      && (r0 = XEXP (PATTERN (insn), 0),
 		  GET_CODE (r0) == REG)
 	      && (link = find_reg_note (insn, REG_LIBCALL, NULL_RTX)) != 0
+	      && XEXP (link, 0) != 0
 	      && GET_CODE (XEXP (link, 0)) == INSN
 	      && (set = single_set (XEXP (link, 0))) != 0
 	      && SET_DEST (set) == r0 && SET_SRC (set) == r0
@@ -1346,7 +1368,7 @@ block_alloc (b)
      decreasing length of life.  Normally call qsort, but if we 
      have only a very small number of quantities, sort them ourselves.  */
 
-  qty_order = (short *) alloca (next_qty * sizeof (short));
+  qty_order = (int *) alloca (next_qty * sizeof (int));
   for (i = 0; i < next_qty; i++)
     qty_order[i] = i;
 
@@ -1376,7 +1398,7 @@ block_alloc (b)
       break;
 
     default:
-      qsort (qty_order, next_qty, sizeof (short), qty_compare_1);
+      qsort (qty_order, next_qty, sizeof (int), qty_compare_1);
     }
 
   /* Try to put each quantity in a suggested physical register, if it has one.
@@ -1476,7 +1498,7 @@ qty_compare (q1, q2)
 
 static int
 qty_compare_1 (q1, q2)
-     short *q1, *q2;
+     int *q1, *q2;
 {
   register int tem;
 
@@ -1981,6 +2003,10 @@ find_free_reg (class, mode, qty, accept_call_clobbered, just_try_suggested,
 			    born_index, dead_index);
     }
 
+  /* We need not check to see if the current function has nonlocal
+     labels because we don't put any pseudos that are live over calls in
+     registers in that case.  */
+
   if (! accept_call_clobbered
       && flag_caller_saves
       && ! just_try_suggested
