@@ -1,5 +1,5 @@
 /* Output variables, constants and external declarations, for GNU compiler.
-   Copyright (C) 1987, 1988, 1989, 1992 Free Software Foundation, Inc.
+   Copyright (C) 1987, 1988, 1989, 1992, 1993 Free Software Foundation, Inc.
 
 This file is part of GNU CC.
 
@@ -333,6 +333,9 @@ make_decl_rtl (decl, asmspec, top_level)
 	error ("function declared `register'");
       else if (DECL_REGISTER (decl) && TYPE_MODE (TREE_TYPE (decl)) == BLKmode)
 	error_with_decl (decl, "data type of `%s' isn't suitable for a register");
+      else if (DECL_REGISTER (decl)
+	       && ! HARD_REGNO_MODE_OK (reg_number, TYPE_MODE (TREE_TYPE (decl))))
+	error_with_decl (decl, "register number for `%s' isn't suitable for the data type");
       /* Now handle properly declared static register variables.  */
       else if (DECL_REGISTER (decl))
 	{
@@ -351,7 +354,16 @@ make_decl_rtl (decl, asmspec, top_level)
 	    error ("global register variable follows a function definition");
 	  if (TREE_THIS_VOLATILE (decl))
 	    warning ("volatile register variables don't work as you might wish");
-	  DECL_RTL (decl) = gen_rtx (REG, DECL_MODE (decl), reg_number);
+
+	  /* If the user specified one of the eliminables registers here,
+	     e.g., FRAME_POINTER_REGNUM, we don't want to get this variable
+	     confused with that register and be eliminated.  Although this
+	     usage is somewhat suspect, we nevertheless use the following
+	     kludge to avoid setting DECL_RTL to frame_pointer_rtx.  */
+
+	  DECL_RTL (decl)
+	    = gen_rtx (REG, DECL_MODE (decl), FIRST_PSEUDO_REGISTER);
+	  REGNO (DECL_RTL (decl)) = reg_number;
 	  REG_USERVAR_P (DECL_RTL (decl)) = 1;
 
 	  if (top_level)
@@ -382,14 +394,17 @@ make_decl_rtl (decl, asmspec, top_level)
 
 	  DECL_RTL (decl) = gen_rtx (MEM, DECL_MODE (decl),
 				     gen_rtx (SYMBOL_REF, Pmode, name));
-	  if (TREE_THIS_VOLATILE (decl))
+	  if (TREE_THIS_VOLATILE (decl)
+	    || (flag_volatile_global && TREE_CODE (decl) == VAR_DECL
+		&& TREE_PUBLIC (decl)))
 	    MEM_VOLATILE_P (DECL_RTL (decl)) = 1;
 	  if (TREE_READONLY (decl))
 	    RTX_UNCHANGING_P (DECL_RTL (decl)) = 1;
 	  MEM_IN_STRUCT_P (DECL_RTL (decl))
 	    = (TREE_CODE (TREE_TYPE (decl)) == ARRAY_TYPE
 	       || TREE_CODE (TREE_TYPE (decl)) == RECORD_TYPE
-	       || TREE_CODE (TREE_TYPE (decl)) == UNION_TYPE);
+	       || TREE_CODE (TREE_TYPE (decl)) == UNION_TYPE
+	       || TREE_CODE (TREE_TYPE (decl)) == QUAL_UNION_TYPE);
 
 	  /* Optionally set flags or add text to the name to record information
 	     such as that it is a function name.
@@ -407,6 +422,7 @@ make_decl_rtl (decl, asmspec, top_level)
 /* Make the rtl for variable VAR be volatile.
    Use this only for static variables.  */
 
+void
 make_var_volatile (var)
      tree var;
 {
@@ -416,6 +432,24 @@ make_var_volatile (var)
   MEM_VOLATILE_P (DECL_RTL (var)) = 1;
 }
 
+/* Output alignment directive to align for constant expression EXP.  */
+
+void
+assemble_constant_align (exp)
+     tree exp;
+{
+  int align;
+
+  /* Align the location counter as required by EXP's data type.  */
+  align = TYPE_ALIGN (TREE_TYPE (exp));
+#ifdef CONSTANT_ALIGNMENT
+  align = CONSTANT_ALIGNMENT (exp, align);
+#endif
+
+  if (align > BITS_PER_UNIT)
+    ASM_OUTPUT_ALIGN (asm_out_file, floor_log2 (align / BITS_PER_UNIT));
+}
+
 /* Output a string of literal assembler code
    for an `asm' keyword used between functions.  */
 
@@ -621,14 +655,15 @@ assemble_zeros (size)
     }
   else
 #endif
-    ASM_OUTPUT_SKIP (asm_out_file, size);
+    if (size > 0)
+      ASM_OUTPUT_SKIP (asm_out_file, size);
 }
 
 /* Assemble a string constant with the specified C string as contents.  */
 
 void
 assemble_string (p, size)
-     unsigned char *p;
+     char *p;
      int size;
 {
   register int i;
@@ -703,8 +738,9 @@ assemble_variable (decl, top_level, at_end)
       return;
     }
 
-  /* Normally no need to say anything for external references,
-     since assembler considers all undefined symbols external.  */
+  /* Normally no need to say anything here for external references,
+     since assemble_external is called by the langauge-specific code
+     when a declaration is first seen.  */
 
   if (DECL_EXTERNAL (decl))
     return;
@@ -1213,6 +1249,16 @@ assemble_real (d, mode)
 
   switch (mode)
     {
+#ifdef ASM_OUTPUT_BYTE_FLOAT
+    case QFmode:
+      ASM_OUTPUT_BYTE_FLOAT (asm_out_file, d);
+      break;
+#endif
+#ifdef ASM_OUTPUT_SHORT_FLOAT
+    case HFmode:
+      ASM_OUTPUT_SHORT_FLOAT (asm_out_file, d);
+      break;
+#endif
 #ifdef ASM_OUTPUT_FLOAT
     case SFmode:
       ASM_OUTPUT_FLOAT (asm_out_file, d);
@@ -1226,6 +1272,7 @@ assemble_real (d, mode)
 #endif
 
 #ifdef ASM_OUTPUT_LONG_DOUBLE
+    case XFmode:
     case TFmode:
       ASM_OUTPUT_LONG_DOUBLE (asm_out_file, d);
       break;
@@ -1265,7 +1312,8 @@ immed_double_const (i0, i1, mode)
   register rtx r;
   int in_current_obstack;
 
-  if (GET_MODE_CLASS (mode) == MODE_INT)
+  if (GET_MODE_CLASS (mode) == MODE_INT
+      || GET_MODE_CLASS (mode) == MODE_PARTIAL_INT)
     {
       /* We clear out all bits that don't belong in MODE, unless they and our
 	 sign bit are all one.  So we get either a reasonable negative value
@@ -1327,10 +1375,10 @@ immed_double_const (i0, i1, mode)
      freed memory.  So switch to saveable_obstack for this allocation
      and then switch back if we were in current_obstack.  */
 
-  in_current_obstack = rtl_in_saveable_obstack ();
+  push_obstacks_nochange ();
+  rtl_in_saveable_obstack ();
   r = gen_rtx (CONST_DOUBLE, mode, 0, i0, i1);
-  if (in_current_obstack)
-    rtl_in_current_obstack ();
+  pop_obstacks ();
 
   /* Don't touch const_double_chain in nested function;
      see force_const_mem.  */
@@ -1370,7 +1418,11 @@ immed_real_const_1 (d, mode)
   /* Avoid REAL_VALUES_EQUAL here in order to distinguish minus zero.  */
   if (!bcmp (&dconst0, &d, sizeof d))
     return CONST0_RTX (mode);
-  else if (REAL_VALUES_EQUAL (dconst1, d))
+  /* Check for NaN first, because some ports (specifically the i386) do not
+     emit correct ieee-fp code by default, and thus will generate a core
+     dump here if we pass a NaN to REAL_VALUES_EQUAL and if REAL_VALUES_EQUAL
+     does a floating point comparison.  */
+  else if (! REAL_VALUE_ISNAN (d) && REAL_VALUES_EQUAL (dconst1, d))
     return CONST1_RTX (mode);
 
   if (sizeof u == 2 * sizeof (HOST_WIDE_INT))
@@ -1396,12 +1448,12 @@ immed_real_const_1 (d, mode)
      freed memory.  So switch to saveable_obstack for this allocation
      and then switch back if we were in current_obstack.  */
 
-  in_current_obstack = rtl_in_saveable_obstack ();
+  push_obstacks_nochange ();
+  rtl_in_saveable_obstack ();
   r = rtx_alloc (CONST_DOUBLE);
   PUT_MODE (r, mode);
   bcopy (&u, &CONST_DOUBLE_LOW (r), sizeof u);
-  if (in_current_obstack)
-    rtl_in_current_obstack ();
+  pop_obstacks ();
 
   /* Don't touch const_double_chain in nested function;
      see force_const_mem.  */
@@ -1788,11 +1840,14 @@ static struct constant_descriptor *
 record_constant (exp)
      tree exp;
 {
-  struct constant_descriptor *ptr = 0;
-  int buf;
+  struct constant_descriptor *next = 0;
+  char *label = 0;
 
-  obstack_grow (&permanent_obstack, &ptr, sizeof ptr);
-  obstack_grow (&permanent_obstack, &buf, sizeof buf);
+  /* Make a struct constant_descriptor.  The first two pointers will
+     be filled in later.  Here we just leave space for them.  */
+
+  obstack_grow (&permanent_obstack, (char *) &next, sizeof next);
+  obstack_grow (&permanent_obstack, (char *) &label, sizeof label);
   record_constant_1 (exp);
   return (struct constant_descriptor *) obstack_finish (&permanent_obstack);
 }
@@ -2554,6 +2609,7 @@ output_constant_pool (fnname, fndecl)
 	  break;
 
 	case MODE_INT:
+	case MODE_PARTIAL_INT:
 	  assemble_integer (x, GET_MODE_SIZE (pool->mode), 1);
 	  break;
 
@@ -2670,6 +2726,8 @@ output_constant (exp, size)
 
   switch (code)
     {
+    case CHAR_TYPE:
+    case BOOLEAN_TYPE:
     case INTEGER_TYPE:
     case ENUMERAL_TYPE:
     case POINTER_TYPE:
